"""
:meta private:
"""

import sympy

from pharmpy.modeling import has_proportional_error_model
from pharmpy.modeling.help_functions import _format_input_list
from pharmpy.parameter import Parameter
from pharmpy.statements import Assignment
from pharmpy.symbols import symbol as S


def set_power_on_ruv(model, list_of_eps=None, ipred=None):
    """Applies a power effect to provided epsilons.

    Initial estimates for new thetas are 1 if the error
    model is proportional, otherwise they are 0.1.

    Parameters
    ----------
    model : Model
        Pharmpy model to create block effect on.
    list_of_eps : str, list
        Name/names of epsilons to apply power effect. If None, all epsilons will be used.
        None is default.
    ipred : Symbol
        Symbol to use as IPRED. Default is to autodetect expression for IPRED.

    Return
    ------
    Model
        Reference to the same model

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> set_power_on_ruv(model)   # doctest: +ELLIPSIS
    <...>
    >>> model.statements.find_assignment("Y")
                 power₁
    Y := EPS(1)⋅F       + F

    See also
    --------
    set_iiv_on_ruv

    """
    list_of_eps = _format_input_list(list_of_eps)
    eps = model.random_variables.epsilons
    if list_of_eps is not None:
        eps = eps[list_of_eps]
    pset, sset = model.parameters, model.statements

    if ipred is None:
        ipred = get_ipred(model)
    else:
        ipred = sympy.sympify(ipred)

    if has_proportional_error_model(model):
        theta_init = 1
    else:
        theta_init = 0.1

    # Find for example W = IPRED
    for s in sset:
        if isinstance(s, Assignment) and s.expression == ipred:
            alternative = s.symbol
            break
    else:
        alternative = None

    for i, e in enumerate(eps):
        theta_name = str(model.create_symbol(stem='power', force_numbering=True))
        theta = Parameter(theta_name, theta_init)
        pset.append(theta)
        sset.subs({e.symbol * ipred: e.symbol})  # To avoid getting F*EPS*F**THETA
        if alternative:  # To avoid getting W*EPS*F**THETA
            sset.subs({e.symbol * alternative: e.symbol})
        sset.subs({e.name: ipred ** S(theta.name) * e.symbol})

    model.parameters = pset
    model.statements = sset

    return model


def get_ipred(model):
    expr = model.statements.after_odes.full_expression(model.dependent_variable)
    for rv in model.random_variables:
        expr = expr.subs(rv.symbol, 0)
    ipred = expr
    for s in model.statements:
        if isinstance(s, Assignment) and s.expression == ipred:
            ipred = s.symbol
            break
    return ipred



import warnings

import numpy as np
import pandas as pd
import sympy

from pharmpy.data_structures import OrderedSet
from pharmpy.model import Model
from pharmpy.modeling import create_rng, sample_parameters_from_covariance_matrix


def calculate_eta_shrinkage(model, sd=False):
    """Calculate eta shrinkage for each eta

    Parameters
    ----------
    model : Model
        Pharmpy model
    sd : bool
        Calculate shrinkage on the standard deviation scale (default is to calculate on the
        variance scale)

    Return
    ------
    Series
        Shrinkage for each eta

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> calculate_eta_shrinkage(model)
    ETA(1)    0.720481
    ETA(2)    0.240295
    dtype: float64
    >>> calculate_eta_shrinkage(model, sd=True)
    ETA(1)    0.471305
    ETA(2)    0.128389
    dtype: float64

    See also
    --------
    calculate_individual_shrinkage

    """
    res = model.modelfit_results
    pe = res.parameter_estimates
    # Want parameter estimates combined with fixed parameter values
    param_inits = model.parameters.to_dataframe()['value']
    pe = pe.combine_first(param_inits)

    ie = res.individual_estimates
    param_names = model.random_variables.iiv.variance_parameters
    diag_ests = pe[param_names]
    diag_ests.index = ie.columns
    if not sd:
        shrinkage = 1 - (ie.var() / diag_ests)
    else:
        shrinkage = 1 - (ie.std() / (diag_ests ** 0.5))
    return shrinkage


def calculate_individual_shrinkage(model):
    """Calculate the individual eta-shrinkage

    Definition: ieta_shr = (var(eta) / omega)

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    DataFrame
        Shrinkage for each eta and individual

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> calculate_individual_shrinkage(model)
          ETA(1)    ETA(2)
    ID
    1   0.847789  0.256473
    2   0.796643  0.210669
    3   0.755025  0.226957
    4   0.764541  0.216405
    5   0.816192  0.203974
    6   0.778108  0.210992
    7   0.659420  0.236875
    8   0.668551  0.240097
    9   0.260056  0.200374
    10  0.725190  0.226563
    11  0.972110  0.421852
    12  0.249640  0.254119
    13  0.730294  0.364932
    14  0.165785  0.194464
    15  0.813399  0.313554
    16  0.797328  0.213211
    17  0.769059  0.278079
    18  0.098506  0.176778
    19  0.749022  0.235386
    20  0.742181  0.222932
    21  0.317956  0.264473
    22  0.943950  0.232732
    23  0.707183  0.259077
    24  0.553787  0.247717
    25  0.826349  0.114302
    26  0.854777  0.341384
    27  0.820829  0.263235
    28  0.999942  0.319986
    29  0.967084  0.432760
    30  0.404773  0.325215
    31  0.999980  0.318421
    32  0.925283  0.167667
    33  0.913706  0.242106
    34  0.875554  0.249197
    35  0.849135  0.294294
    36  0.172206  0.246422
    37  0.747380  0.278340
    38  0.187440  0.231249
    39  0.237805  0.254485
    40  0.999925  0.189793
    41  0.941906  0.170998
    42  0.923801  0.244046
    43  0.999928  0.320403
    44  0.237637  0.260453
    45  0.869540  0.194503
    46  0.999949  0.319750
    47  0.983782  0.393234
    48  0.698267  0.169337
    49  0.776674  0.214962
    50  0.688847  0.192608
    51  0.822213  0.202534
    52  0.511489  0.273601
    53  0.964757  0.223448
    54  0.762153  0.181648
    55  0.965657  0.435741
    56  0.995278  0.354798
    57  0.813382  0.263372
    58  0.727295  0.232867
    59  0.738777  0.224742

    See also
    --------
    calculate_eta_shrinkage

    """
    res = model.modelfit_results
    cov = res.individual_estimates_covariance
    pe = res.parameter_estimates
    # Want parameter estimates combined with fixed parameter values
    param_inits = model.parameters.to_dataframe()['value']
    pe = pe.combine_first(param_inits)

    # Get all iiv variance parameters
    param_names = model.random_variables.etas.variance_parameters
    param_names = list(OrderedSet(param_names))  # Only unique in order

    diag_ests = pe[param_names]

    def fn(row, ests):
        names = row[0].index
        ser = pd.Series(np.diag(row[0].values) / ests, index=names)
        return ser

    ish = pd.DataFrame(cov).apply(fn, axis=1, ests=diag_ests.values)
    return ish


def calculate_individual_parameter_statistics(model, exprs, rng=None):
    """Calculate statistics for individual parameters

    Calculate the mean (expected value of the distribution), variance
    (variance of the distribution) and standard error for individual
    parameters described by arbitrary expressions. Any dataset column or
    variable used in the model can be used in the expression. The exception
    being that variables that depends on the solution of the ODE system
    cannot be used. If covariates are used in the expression the statistics
    of the parameter is calculated at the median value of each covariate as well
    as at the 5:th and 95:th percentiles. If no parameter uncertainty is available
    for the model the standard error will not be calculated.

    Parameters
    ----------
    model : Model
        A previously estimated model
    exprs : str, sympy expression or iterable of str or sympy expressions
        Expressions or equations for parameters of interest. If equations are used
        the names of the left hand sides will be used as the names of the parameters.
    rng : Generator or int
        Random number generator or int seed

    Returns
    -------
    pd.DataFrame
        A DataFrame of statistics indexed on parameter and covariate value.

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, create_rng
    >>> from pharmpy.modeling import calculate_individual_parameter_statistics
    >>> model = load_example_model("pheno")
    >>> rng = create_rng(23)
    >>> calculate_individual_parameter_statistics(model, "K=CL/V", rng=rng)
                              mean  variance    stderr
    parameter covariates
    K         p5          0.004234  0.000001  0.001138
              median      0.004909  0.000001  0.001212
              p95         0.004910  0.000001  0.001263
    """
    rng = create_rng(rng)

    if isinstance(exprs, str) or isinstance(exprs, sympy.Basic):
        exprs = [_split_equation(exprs)]
    else:
        exprs = [_split_equation(expr) for expr in exprs]
    dataset = model.dataset
    cols = set(dataset.columns)
    i = 0
    table = pd.DataFrame(columns=['parameter', 'covariates', 'mean', 'variance', 'stderr'])
    for name, expr in exprs:
        full_expr = model.statements.before_odes.full_expression(expr)
        covariates = {symb.name for symb in full_expr.free_symbols if symb.name in cols}
        if not covariates:
            cases = {'median': dict()}
        else:
            q5 = dataset[{'ID'} | covariates].groupby('ID').median().quantile(0.05)
            q95 = dataset[{'ID'} | covariates].groupby('ID').median().quantile(0.95)
            median = dataset[{'ID'} | covariates].groupby('ID').median().median()
            cases = {'p5': dict(q5), 'median': dict(median), 'p95': dict(q95)}

        df = pd.DataFrame(index=list(cases.keys()), columns=['mean', 'variance', 'stderr'])
        for case, cov_values in cases.items():
            pe = dict(model.modelfit_results.parameter_estimates)
            cov_expr = full_expr.subs(cov_values)
            expr = cov_expr.subs(pe)
            samples = model.random_variables.sample(expr, parameters=pe, samples=1000000, rng=rng)

            mean = np.mean(samples)
            variance = np.var(samples)

            if model.modelfit_results.covariance_matrix is not None:
                parameters = sample_parameters_from_covariance_matrix(
                    model,
                    n=100,
                    force_posdef_covmatrix=True,
                    rng=rng,
                )
                samples = []
                with warnings.catch_warnings():
                    warnings.filterwarnings('ignore')
                    for _, row in parameters.iterrows():
                        batch = model.random_variables.sample(
                            cov_expr.subs(dict(row)),
                            parameters=dict(row),
                            samples=10,
                            rng=rng,
                        )
                        samples.extend(list(batch))
                stderr = pd.Series(samples).std()
            else:
                stderr = np.nan
            df.loc[case] = [mean, variance, stderr]
            df.index.name = 'covariates'

        df.reset_index(inplace=True)
        if not name:
            name = f'unknown{i}'
            i += 1
        df['parameter'] = name
        table = pd.concat([table, df])
    table.set_index(['parameter', 'covariates'], inplace=True)
    table = table.astype('float64')
    return table


def calculate_pk_parameters_statistics(model, rng=None):
    """Calculate statistics for common pharmacokinetic parameters

    Calculate the mean (expected value of the distribution), variance
    (variance of the distribution) and standard error for some individual
    pre-defined pharmacokinetic parameters.

    Parameters
    ----------
    model : Model
        A previously estimated model
    rng : Generator or int
        Random number generator or seed

    Returns
    -------
    pd.DataFrame
        A DataFrame of statistics indexed on parameter and covariate value.

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, create_rng
    >>> from pharmpy.modeling import calculate_pk_parameters_statistics
    >>> model = load_example_model("pheno")
    >>> rng = create_rng(23)
    >>> calculate_pk_parameters_statistics(model, rng=rng)  # doctest: +NORMALIZE_WHITESPACE
                              mean  variance    stderr
    parameter covariates
    k_e       p5          0.004234  0.000001  0.001138
              median      0.004909  0.000001  0.001212
              p95         0.004910  0.000001  0.001263

    See Also
    --------
    calculate_individual_parameter_statistics : Calculation of statistics for arbitrary parameters

    """

    statements = model.statements
    odes = statements.ode_system
    central = odes.central_compartment
    output = odes.output_compartment
    depot = odes.find_depot(statements)
    peripherals = odes.peripheral_compartments
    elimination_rate = odes.get_flow(central, output)

    expressions = []  # Eq(name, expr)
    # FO abs + 1comp + FO elimination
    if len(odes) == 3 and depot and odes.t not in elimination_rate.free_symbols:
        exodes = odes.to_explicit_system(skip_output=True)
        ode_list, ics = exodes.odes, exodes.ics
        sols = sympy.dsolve(ode_list, ics=ics)
        expr = sols[1].rhs
        d = sympy.diff(expr, odes.t)
        tmax_closed_form = sympy.solve(d, odes.t)[0]
        expressions.append(sympy.Eq(sympy.Symbol('t_max'), tmax_closed_form))
        e2 = sympy.simplify(expr / depot.dose.amount / sympy.denom(elimination_rate))
        cmax_dose_closed_form = sympy.simplify(e2.subs({odes.t: tmax_closed_form}))
        expressions.append(sympy.Eq(sympy.Symbol('C_max_dose'), cmax_dose_closed_form))

    # Any abs + 1comp + FO elimination
    if not peripherals and odes.t not in elimination_rate.free_symbols:
        elimination_system = statements.copy().ode_system
        # keep central and output
        for name in elimination_system.names:
            if name not in [central.name, output.name]:
                elimination_system.remove_compartment(elimination_system.find_compartment(name))
                exodes = elimination_system.to_explicit_system(skip_output=True)
                ode_list, ics = exodes.odes, exodes.ics
                A0 = sympy.Symbol('A0')
                ic = ics.popitem()
                ics = {ic[0]: A0}
                sols = sympy.dsolve(ode_list[0], ics=ics)
                eq = sympy.Eq(sympy.Rational(1, 2) * A0, sols.rhs)
                thalf_elim = sympy.solve(eq, odes.t)[0]
                expressions.append(sympy.Eq(sympy.Symbol('t_half_elim'), thalf_elim))

    # Bolus dose + 2comp + FO elimination
    if len(peripherals) == 1 and len(odes) == 3 and odes.t not in elimination_rate.free_symbols:
        exodes = odes.to_explicit_system(skip_output=True)
        ode_list, ics = exodes.odes, exodes.ics
        sols = sympy.dsolve(ode_list, ics=ics)
        A = sympy.Wild('A')
        B = sympy.Wild('B')
        alpha = sympy.Wild('alpha')
        beta = sympy.Wild('beta')

        m = sols[0].rhs.match(A * sympy.exp(-alpha) + B * sympy.exp(-beta))

        beta = m[beta] / odes.t
        alpha = m[alpha] / odes.t
        A = m[A] / central.dose.amount
        B = m[B] / central.dose.amount

        if (alpha - alpha).extract_multiplicatively(-1) is not None:
            # alpha > beta  (sympy couldn't simplify this directly)
            alpha, beta = beta, alpha
            A, B = B, A
        expressions.append(sympy.Eq(sympy.Symbol('A'), A))
        expressions.append(sympy.Eq(sympy.Symbol('B'), B))
        expressions.append(sympy.Eq(sympy.Symbol('alpha'), alpha))
        expressions.append(sympy.Eq(sympy.Symbol('beta'), beta))

    # Any abs + any comp + FO elimination
    if odes.t not in elimination_rate.free_symbols:
        expressions.append(sympy.Eq(sympy.Symbol('k_e'), elimination_rate))

    df = calculate_individual_parameter_statistics(model, expressions, rng=rng)
    return df


def _split_equation(s):
    if isinstance(s, str):
        a = s.split('=')
        if len(a) == 1:
            name = None
            expr = sympy.sympify(s)
        else:
            name = a[0].strip()
            expr = sympy.sympify(a[1])
    elif isinstance(s, sympy.Eq):
        name = s.lhs.name
        expr = s.rhs
    else:  # sympy expr
        name = None
        expr = s
    if name is None and isinstance(expr, sympy.Symbol):
        name = expr.name
    return name, expr


def summarize_modelfit_results(models, include_all_estimation_steps=False):
    """Summarize results of model runs

    Summarize different results after fitting a model, includes runtime, ofv,
    and parameter estimates (with errors). If include_all_estimation_steps is False,
    only the last estimation step will be included (note that in that case, the
    minimization_successful value will be referring to the last estimation step, if
    last step is evaluation it will go backwards until it finds an estimation step
    that wasn't an evaluation).

    Parameters
    ----------
    models : list, Model
        List of models or single model
    include_all_estimation_steps : bool
        Whether to include all estimation steps, default is False

    Return
    ------
    pd.DataFrame
        A DataFrame of modelfit results with model name and estmation step as index.

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, summarize_modelfit_results
    >>> model = load_example_model("pheno")
    >>> summarize_modelfit_results([model]) # doctest: +ELLIPSIS
          minimization_successful         ofv runtime_total  THETA(1)_estimate  ... SIGMA(1,1)_RSE
    pheno                    True  586.276056           4.0           0.004696  ...       0.172147
    """
    if isinstance(models, Model):
        models = [models]

    summaries = []

    for model in models:
        res = model.modelfit_results
        if res:
            summaries.append(res.result_summary(include_all_estimation_steps))
        else:
            # FIXME: in read_modelfit_results, maybe some parts can be extracted (i.e.
            #   create modelfit_results object)
            if include_all_estimation_steps:
                for i in range(len(model.estimation_steps)):
                    index = pd.MultiIndex.from_tuples(
                        [(model.name, i + 1)], names=['model_name', 'step']
                    )
                    empty_df = pd.DataFrame(
                        {'run_type': None, 'minimization_successful': None}, index=index
                    )
                    summaries.append(empty_df)
            else:
                empty_df = pd.DataFrame({'minimization_successful': np.nan}, index=[model.name])
                summaries.append(empty_df)

    summary = pd.concat(summaries).sort_index()

    return summary



import shutil

from pharmpy import Model
from pharmpy.tools.bootstrap import Bootstrap
from pharmpy.utils import TemporaryDirectoryChanger


def test_bootstrap(tmp_path, testdata):
    with TemporaryDirectoryChanger(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'pheno.mod', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'pheno.dta', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'pheno.ext', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'pheno.lst', tmp_path)
        model = Model('pheno.mod')
        model.dataset_path = tmp_path / 'pheno.dta'
        model.modelfit_results.ofv  # Read in results
        res = Bootstrap(model, 3).run()
        assert len(res.parameter_estimates) == 3



from pharmpy import Model


def test_data_filename_set(testdata):
    model = Model(testdata / 'nonmem' / 'pheno_abbr.mod')
    assert model.control_stream.abbreviated.replace == {
        'THETA(CL)': 'THETA(1)',
        'THETA(V)': 'THETA(2)',
        'ETA(CL)': 'ETA(1)',
        'ETA(V)': 'ETA(2)',
    }


def test_replace(parser):
    rec = parser.parse('$ABBR REPLACE K34="3,4"').records[0]
    assert rec.replace == {'K34': '3,4'}


def test_translate_to_pharmpy_names(testdata):
    model = Model(testdata / 'nonmem' / 'pheno_abbr.mod')
    new_dict = model.control_stream.abbreviated.translate_to_pharmpy_names()
    assert new_dict == {
        'THETA(1)': 'THETA_CL',
        'THETA(2)': 'THETA_V',
        'ETA(1)': 'ETA_CL',
        'ETA(2)': 'ETA_V',
    }



"""
.. highlight:: console

============================
The CLI interface of Pharmpy
============================

Examples
--------

Example argument lines (only preceeded by ``pharmpy`` or ``python3 -m pharmpy``)::

    # filter data (as model prescribes) and output to file 'output.csv'
    pharmpy data write run1.mod -o output.csv

Entrypoints and Subcommands
---------------------------

Standard practice of packaging is to define "entrypoints" in the setuptools ``setup.py``. Currently:

.. code-block:: python

  entry_points={
      'console_scripts': [
          'pharmpy           = pharmpy.cli:main',
      ]
  },

Main entrypoint is ``pharmpy`` (which maps :func:`main`), but each subcommand could be given a
shortcutting entrypoint (e.g. :func:`sumo`). This would be accomplished via invoking:

Installing Pharmpy Wheel-build should install all of these to PATH. Then, these are clean targets
for post-install hooks to symlink. Either to pollute the PATH namespace with traditional PsN-styled
"binscripts", as::

    /usr/bin/execute               ->  /usr/bin/pharmpy execute

Or, to keep multiple versions when upgrading Pharmpy.

On Logging
----------

As a CLI we are at the top, interacting with the user. All of Pharmpy will get messages
filtered through here (loggers organize themselves in a hiearchy based on their period-split names).

.. warning::
    Loggers **mustn't** be configurated if imported as library (e.g. set handlers and such),
    since that should bubble up into the lap of whomever is at the top (of the Python
    period-separated namespace).

Definitions
===========
"""

import argparse
import importlib
import pathlib
import pydoc
import sys
import types
import warnings
from collections import OrderedDict, namedtuple
from textwrap import dedent


class LazyLoader(types.ModuleType):
    """Class that masquerades as a module and lazily loads it when accessed the first time

    The code for the class is taken from TensorFlow and is under the Apache 2.0 license

    This is needed for the cli to be able to do things as version and help quicker than
    if all modules were imported.
    """

    def __init__(self, local_name, parent_module_globals, name):
        self._local_name = local_name
        self._parent_module_globals = parent_module_globals

        super(LazyLoader, self).__init__(name)

    def _load(self):
        # Import the target module and insert it into the parent's namespace
        module = importlib.import_module(self.__name__)
        self._parent_module_globals[self._local_name] = module

        # Update this object's dict so that if someone keeps a reference to the
        #   LazyLoader, lookups are efficient (__getattr__ is only called on lookups
        #   that fail).
        self.__dict__.update(module.__dict__)
        return module

    def __getattr__(self, item):
        module = self._load()
        return getattr(module, item)

    def __dir__(self):
        module = self._load()
        return dir(module)


pharmpy = LazyLoader('pharmpy', globals(), 'pharmpy')
np = LazyLoader('numpy', globals(), 'numpy')
iterators = LazyLoader('iterators', globals(), 'pharmpy.data.iterators')
plugin_utils = LazyLoader('plugin_utils', globals(), 'pharmpy.plugins.utils')

formatter = argparse.ArgumentDefaultsHelpFormatter


def warnings_formatter(message, *args, **kwargs):
    return f'Warning: {message}\n'


warnings.formatwarning = warnings_formatter


def error(exception):
    """Raise the exception with no traceback printed

    Used for non-recoverables in CLI. Exceptions giving traceback are bugs!
    """

    def exc_only_hook(exception_type, exception, traceback):
        print(f'{exception_type.__name__}: {exception}')

    sys.excepthook = exc_only_hook
    raise exception


def format_keyval_pairs(data_dict, sort=True, right_just=False):
    """Formats lines from *data_dict*."""

    if not data_dict:
        return []

    key_width = max(len(field) for field in data_dict.keys())
    if sort:
        data_dict = OrderedDict(sorted(data_dict.items()))
    if right_just:
        line_format = '  %%%ds\t%%s' % key_width
    else:
        line_format = '%%-%ds\t%%s' % key_width

    lines = []
    for key, values in data_dict.items():
        if isinstance(values, str):
            values = values.splitlines()
        keys = [key] + [''] * (len(values) - 1)
        for k, v in zip(keys, values):
            lines += [line_format % (k, v)]

    return lines


def run_bootstrap(args):
    import pharmpy.tools.bootstrap as bootstrap

    method = bootstrap.Bootstrap(args.model, resamples=args.samples)
    method.run()


def run_execute(args):
    import pharmpy.tools.modelfit as modelfit

    method = modelfit.Modelfit(args.models)
    method.run()


def run_modelsearch(args):
    import pharmpy.tools.modelsearch as modelsearch

    method = modelsearch.ModelSearch(args.model, args.algorithm, args.funcs, rankfunc=args.rankfunc)
    method.run()


def data_write(args):
    """Subcommand to write a dataset."""
    try:
        df = args.model_or_dataset.dataset
    except plugin_utils.PluginError:
        df = args.model_or_dataset
    path = args.output_file
    try:
        # If no output_file supplied will use name of df
        path = df.pharmpy.write_csv(path=path, force=args.force)
        print(f'Dataset written to {path}')
    except OSError as e:
        error(e)


def data_print(args):
    """Subcommand to print a dataset to stdout via pager"""
    try:
        df = args.model_or_dataset.dataset
    except AttributeError:
        df = args.model_or_dataset
    if args.columns:
        try:
            df = df[args.columns]
        except KeyError as e:
            error(e)
    pydoc.pager(df.to_string())


def data_filter(args):
    """Subcommand to filter a dataset"""
    try:
        df = args.model_or_dataset.dataset
    except plugin_utils.PluginError:
        df = args.model_or_dataset
    expression = ' '.join(args.expressions)
    try:
        df.query(expression, inplace=True)
    except SyntaxError:
        error(SyntaxError(f'Invalid syntax of query: "{expression}"'))
    except BaseException as e:
        error(e)
    write_model_or_dataset(args.model_or_dataset, df, args.output_file, args.force)


def data_append(args):
    """Subcommand to append a column to a dataset"""
    try:
        df = args.model_or_dataset.dataset
    except plugin_utils.PluginError:
        df = args.model_or_dataset
    try:
        df.eval(args.expression, inplace=True)
    except SyntaxError:
        error(SyntaxError(f'Invalid syntax of expression: "{args.expression}"'))
    write_model_or_dataset(args.model_or_dataset, df, args.output_file, args.force)


def write_model_or_dataset(model_or_dataset, new_df, path, force):
    """Write model or dataset to output_path or using default names"""
    if hasattr(model_or_dataset, 'pharmpy'):
        # Is a dataset
        try:
            # If no output_file supplied will use name of df
            path = new_df.pharmpy.write_csv(path=path, force=force)
            print(f'Dataset written to {path}')
        except FileExistsError as e:
            error(
                exception=FileExistsError(
                    f'{e.args[0]} Use -f  or --force to ' 'force an overwrite'
                )
            )
    else:
        # Is a model
        model = model_or_dataset
        if new_df is not None:
            model.dataset = new_df
        try:
            if path:
                if not path.is_dir():
                    model.bump_model_number(path)
                model.write(path=path, force=force)
            else:
                model.bump_model_number()
                model.write(force=force)
        except FileExistsError as e:
            error(FileExistsError(f'{e.args[0]} Use -f or --force to ' 'force an overwrite'))


def data_resample(args):
    """Subcommand to resample a dataset."""
    resampler = iterators.Resample(
        args.model_or_dataset,
        args.group,
        resamples=args.resamples,
        stratify=args.stratify,
        replace=args.replace,
        sample_size=args.sample_size,
        name_pattern=f'{args.model_or_dataset.name}_resample_{{}}',
    )
    for resampled_obj, _ in resampler:
        try:
            try:
                resampled_obj.write()
            except AttributeError:
                resampled_obj.pharmpy.write_csv()
        except OSError as e:
            error(e)


def data_anonymize(args):
    """Subcommand to anonymize a dataset
    anonymization is really a special resample case where the specified group
    (usually the ID) is given new random unique values and reordering.
    Will be default try to overwrite original file (but won't unless force)
    """
    path = check_input_path(args.dataset)
    dataset = pharmpy.data.read_csv(path, raw=True, parse_columns=[args.group])
    resampler = pharmpy.data.iterators.Resample(dataset, args.group)
    df, _ = next(resampler)
    if args.output_file:
        output_file = args.output_file
    else:
        output_file = path
    write_model_or_dataset(df, df, output_file, args.force)


def info(args):
    """Subcommand to print Pharmpy info (and brag a little bit)."""

    Install = namedtuple('VersionInfo', ['version', 'authors', 'directory'])
    inst = Install(
        pharmpy.__version__,
        'A list of authors can be found in the AUTHORS.rst',
        str(pathlib.Path(pharmpy.__file__).parent),
    )

    lines = format_keyval_pairs(inst._asdict(), right_just=True)
    print('\n'.join(lines))


def model_print(args):
    """Subcommand for formatting/printing model components."""
    lines = []
    for i, model in enumerate(args.models):
        lines += ['[%d/%d] %r' % (i + 1, len(args.models), model.name)]
        dict_ = OrderedDict()
        try:
            dict_['dataset'] = repr(model.dataset)
        except FileNotFoundError as e:
            dict_['dataset'] = str(e)
        s = repr(model.parameters) + '\n\n'
        dict_['parameters'] = s
        s = repr(model.random_variables) + '\n\n'
        dict_['random_variables'] = s
        s = str(model.statements) + '\n\n'
        dict_['model_statements'] = s
        dict_lines = format_keyval_pairs(dict_, sort=False)
        lines += ['\t%s' % line for line in dict_lines]
    if len(lines) > 24:
        pydoc.pager('\n'.join(lines))
    else:
        print('\n'.join(lines))


def model_sample(args):
    model = args.model
    from pharmpy.modeling import sample_parameters_from_covariance_matrix

    samples = sample_parameters_from_covariance_matrix(model, n=args.samples)
    for row, params in samples.iterrows():
        model.parameters = params
        model.name = f'sample_{row + 1}'
        model.write()


def add_covariate_effect(args):
    """Subcommand to add covariate effect to model."""
    from pharmpy.modeling import add_covariate_effect

    model = args.model
    add_covariate_effect(model, args.param, args.covariate, args.effect, args.operation)

    write_model_or_dataset(model, model.dataset, path=args.output_file, force=False)


def model_absorption_rate(args):
    import pharmpy.modeling as modeling

    model = args.model
    if args.order == 'ZO':
        modeling.set_zero_order_absorption(model)
    elif args.order == 'FO':
        modeling.set_first_order_absorption(model)
    elif args.order == 'bolus':
        modeling.set_bolus_absorption(model)
    elif args.order == 'seq-ZO-FO':
        modeling.set_seq_zo_fo_absorption(model)
    write_model_or_dataset(model, None, path=args.output_file, force=args.force)


def model_elimination_rate(args):
    import pharmpy.modeling as modeling

    model = args.model
    if args.order == 'ZO':
        modeling.set_zero_order_elimination(model)
    elif args.order == 'FO':
        modeling.set_first_order_elimination(model)
    elif args.order == 'MM':
        modeling.set_michaelis_menten_elimination(model)
    elif args.order == 'comb-FO-MM':
        modeling.combined_mm_fo_elimination(model)
    write_model_or_dataset(model, None, path=args.output_file, force=args.force)


def model_transit_compartments(args):
    from pharmpy.modeling import set_transit_compartments

    model = args.model
    set_transit_compartments(model, args.n)
    write_model_or_dataset(model, None, path=args.output_file, force=args.force)


def model_peripheral_compartments(args):
    from pharmpy.modeling import add_peripheral_compartment, remove_peripheral_compartment

    model = args.model
    p = model.statements.ode_system.peripheral_compartments
    if len(p) < args.n:
        func = add_peripheral_compartment
    elif len(p) > args.n:
        func = remove_peripheral_compartment
    else:
        return
    iters = abs(len(p) - args.n)
    for _ in range(iters):
        func(model)
    write_model_or_dataset(model, None, path=args.output_file, force=args.force)


def model_error(args):
    from pharmpy.modeling import error_model

    model = args.model
    error_model(model, args.error_type)
    write_model_or_dataset(model, None, path=args.output_file, force=args.force)


def boxcox(args):
    """Subcommand to apply boxcox transformation to specified etas of model."""
    from pharmpy.modeling import transform_etas_boxcox

    model = args.model
    try:
        etas = args.etas.split(" ")
    except AttributeError:
        etas = args.etas

    transform_etas_boxcox(model, etas)
    write_model_or_dataset(model, None, path=args.output_file, force=False)


def tdist(args):
    """Subcommand to apply t-distribution transformation to specified etas of model."""
    from pharmpy.modeling import transform_etas_tdist

    model = args.model
    try:
        etas = args.etas.split(" ")
    except AttributeError:
        etas = args.etas

    transform_etas_tdist(model, etas)
    write_model_or_dataset(model, None, path=args.output_file, force=False)


def john_draper(args):
    """Subcommand to apply John Draper transformation to specified etas of model."""
    from pharmpy.modeling import transform_etas_john_draper

    model = args.model
    try:
        etas = args.etas.split(" ")
    except AttributeError:
        etas = args.etas

    transform_etas_john_draper(model, etas)
    write_model_or_dataset(model, None, path=args.output_file, force=False)


def add_iiv(args):
    """Subcommand to add new IIVs to model."""
    from pharmpy.modeling import add_iiv

    model = args.model
    add_iiv(model, args.param, args.expression, args.operation, args.eta_name)

    write_model_or_dataset(model, model.dataset, path=args.output_file, force=False)


def add_iov(args):
    """Subcommand to add new IOVs to model."""
    from pharmpy.modeling import add_iov

    model = args.model
    try:
        etas = args.etas.split(" ")
    except AttributeError:
        etas = args.etas

    try:
        eta_names = args.eta_names.split(" ")
    except AttributeError:
        eta_names = args.eta_names

    add_iov(model, args.occ, etas, eta_names)

    write_model_or_dataset(model, model.dataset, path=args.output_file, force=False)


def create_joint_distribution(args):
    """Subcommand to create full or partial block structures."""
    from pharmpy.modeling import create_joint_distribution

    model = args.model

    try:
        etas = args.etas.split(" ")
    except AttributeError:
        etas = args.etas

    create_joint_distribution(model, etas)

    write_model_or_dataset(model, model.dataset, path=args.output_file, force=False)


def iiv_on_ruv(args):
    """Subcommand to multiply epsilons with exponential etas."""
    from pharmpy.modeling import set_iiv_on_ruv

    model = args.model

    try:
        eps = args.eps.split(" ")
    except AttributeError:
        eps = args.eps

    try:
        eta_names = args.eta_names.split(" ")
    except AttributeError:
        eta_names = args.eta_names

    set_iiv_on_ruv(model, eps, args.same_eta, eta_names)

    write_model_or_dataset(model, model.dataset, path=args.output_file, force=False)


def remove_iiv(args):
    """Subcommand to remove IIVs."""
    from pharmpy.modeling import remove_iiv

    model = args.model

    try:
        to_remove = args.to_remove.split(" ")
    except AttributeError:
        to_remove = args.to_remove

    remove_iiv(model, to_remove)

    write_model_or_dataset(model, model.dataset, path=args.output_file, force=False)


def remove_iov(args):
    """Subcommand to remove IOVs."""
    from pharmpy.modeling import remove_iov

    model = args.model
    remove_iov(model)

    write_model_or_dataset(model, model.dataset, path=args.output_file, force=False)


def power_on_ruv(args):
    """Subcommand to apply power effect to RUVs."""
    from pharmpy.modeling import set_power_on_ruv

    model = args.model

    try:
        eps = args.eps.split(" ")
    except AttributeError:
        eps = args.eps

    set_power_on_ruv(model, eps)

    write_model_or_dataset(model, model.dataset, path=args.output_file, force=False)


def update_inits(args):
    """Subcommand to update initial estimates from previous output."""
    from pharmpy.modeling import update_inits

    model = args.model
    update_inits(model, args.force_update)

    write_model_or_dataset(model, model.dataset, path=args.output_file, force=False)


def results_bootstrap(args):
    """Subcommand to generate bootstrap results"""
    from pharmpy.tools.bootstrap.results import psn_bootstrap_results

    if not args.psn_dir.is_dir():
        error(FileNotFoundError(str(args.psn_dir)))
    try:
        res = psn_bootstrap_results(args.psn_dir)
    except FileNotFoundError as e:
        error(e)
    res.to_json(path=args.psn_dir / 'results.json')
    res.to_csv(path=args.psn_dir / 'results.csv')


def results_cdd(args):
    from pharmpy.tools.cdd.results import psn_cdd_results

    if not args.psn_dir.is_dir():
        error(FileNotFoundError(str(args.psn_dir)))
    res = psn_cdd_results(args.psn_dir)
    res.to_json(path=args.psn_dir / 'results.json')
    res.to_csv(path=args.psn_dir / 'results.csv')


def results_frem(args):
    """Generate frem results"""
    from pharmpy.tools.frem.results import psn_frem_results

    if not args.psn_dir.is_dir():
        error(FileNotFoundError(str(args.psn_dir)))
    res = psn_frem_results(
        args.psn_dir,
        method=args.method,
        force_posdef_covmatrix=args.force_posdef_covmatrix,
        force_posdef_samples=args.force_posdef_samples,
    )
    res.add_plots()
    res.to_json(path=args.psn_dir / 'results.json')
    res.to_csv(path=args.psn_dir / 'results.csv')
    res.create_report(args.psn_dir)


def results_linearize(args):
    from pharmpy.tools.linearize.results import psn_linearize_results

    if not args.psn_dir.is_dir():
        error(FileNotFoundError(str(args.psn_dir)))
    res = psn_linearize_results(args.psn_dir)
    res.to_json(path=args.psn_dir / 'results.json')
    res.to_csv(path=args.psn_dir / 'results.csv')


def results_resmod(args):
    from pharmpy.tools.resmod.results import psn_resmod_results

    if not args.psn_dir.is_dir():
        error(FileNotFoundError(str(args.psn_dir)))
    res = psn_resmod_results(args.psn_dir)
    res.to_json(path=args.psn_dir / 'results.json')
    res.to_csv(path=args.psn_dir / 'results.csv')


def results_scm(args):
    from pharmpy.tools.scm.results import psn_scm_results

    if not args.psn_dir.is_dir():
        error(FileNotFoundError(str(args.psn_dir)))
    res = psn_scm_results(args.psn_dir)
    res.to_json(path=args.psn_dir / 'results.json')
    res.to_csv(path=args.psn_dir / 'results.csv')


def results_simeval(args):
    from pharmpy.tools.simeval.results import psn_simeval_results

    if not args.psn_dir.is_dir():
        error(FileNotFoundError(str(args.psn_dir)))
    res = psn_simeval_results(args.psn_dir)
    res.to_json(path=args.psn_dir / 'results.json')
    res.to_csv(path=args.psn_dir / 'results.csv')


def results_ofv(args):
    """Subcommand to extract final ofv from multiple results"""
    ofvs = []
    for model in args.models:
        try:
            ofv = str(model.modelfit_results.ofv)
        except AttributeError:
            ofv = 'NA'
        ofvs.append(ofv)
    print(' '.join(ofvs))


def results_print(args):
    """Subcommand to print any results"""
    if args.dir.is_dir():
        path = args.dir / 'results.json'
    elif args.dir.is_file():
        path = args.dir
    else:
        error(FileNotFoundError(str(args.dir)))
    from pharmpy.results import read_results

    res = read_results(path)
    print(res)


def results_qa(args):
    from pharmpy.tools.qa.results import psn_qa_results

    if not args.psn_dir.is_dir():
        error(FileNotFoundError(str(args.psn_dir)))
    res = psn_qa_results(args.psn_dir)
    res.to_json(path=args.psn_dir / 'results.json')
    res.to_csv(path=args.psn_dir / 'results.csv')


def results_report(args):
    """Subcommand to generate reports"""
    results_path = args.psn_dir / 'results.json'
    if not results_path.is_file():
        error(FileNotFoundError(results_path))
    from pharmpy.results import read_results

    res = read_results(results_path)
    res.create_report(args.psn_dir)


def results_summary(args):
    """Subcommand to output summary of modelfit"""
    for model in args.models:
        print(model.name)
        print(model.modelfit_results.parameter_summary())
        print()


def check_input_path(path):
    """Resolves path to input file and checks existence.

    Raises if not found or is dir, without tracebacks (see :func:`error_exit`).
    """
    try:
        path = pathlib.Path(path)
        path = path.resolve()
    except FileNotFoundError:
        pass

    if not path.exists():
        exc = FileNotFoundError('No such input file: %r' % str(path))
        error(exc)
    elif path.is_dir():
        exc = IsADirectoryError('Is a directory (not an input file): %r' % str(path))
        error(exc)
    else:
        return path


# ------ Type functions --------------------


def input_model(path):
    """Returns :class:`~pharmpy.model.Model` from *path*.

    Raises if not found or is dir, without tracebacks (see :func:`error_exit`).
    """
    path = check_input_path(path)
    return pharmpy.Model(path)


def input_model_or_dataset(path):
    """Returns :class:`~pharmpy.model.Model` or :class:`~pharmpy.data.PharmDataFrame` from *path*"""
    path = check_input_path(path)
    try:
        obj = pharmpy.Model(path)
    except pharmpy.plugins.utils.PluginError:
        obj = pharmpy.data.read_csv(path)
    return obj


def comma_list(s):
    if s == '':
        return []
    return s.split(',')


def semicolon_list(s):
    if s == '':
        return []
    return s.split(';')


def random_seed(seed):
    try:
        seed = int(seed)
        np.random.seed(seed)
    except ValueError as e:
        # Cannot reraise the ValueError as it would be caught by argparse
        error(argparse.ArgumentTypeError(str(e)))
    return seed


# for commands taking multiple model files as input
args_input = argparse.ArgumentParser(add_help=False)
group_input = args_input.add_argument_group(title='inputs')
group_input.add_argument(
    'models', metavar='FILE', type=input_model, nargs='+', help='input model files'
)

# for commands taking one model file as input
args_model_input = argparse.ArgumentParser(add_help=False)
group_model_input = args_model_input.add_argument_group(title='model')
group_model_input.add_argument('model', metavar='FILE', type=input_model, help='input model file')

# for commands with model file or dataset input
args_model_or_data_input = argparse.ArgumentParser(add_help=False)
group_model_or_data_input = args_model_or_data_input.add_argument_group(title='data_input')
group_model_or_data_input.add_argument(
    'model_or_dataset',
    metavar='FILE',
    type=input_model_or_dataset,
    help='input model or dataset file',
)

# for commands involving randomization
args_random = argparse.ArgumentParser(add_help=False)
group_random = args_random.add_argument_group(title='random seed')
group_random.add_argument(
    '--seed',
    type=random_seed,
    metavar='INTEGER',
    help='Provide a random seed. The seed must be an integer ' 'between 0 and 2^32 - 1',
)

# for commands with file output
args_output = argparse.ArgumentParser(add_help=False)
group_output = args_output.add_argument_group(title='outputs')
group_output.add_argument(
    '-f', '--force', action='store_true', help='remove existing destination files (all)'
)
group_output.add_argument(
    '-o', '--output_file', dest='output_file', metavar='file', type=pathlib.Path, help='output file'
)

parser_definition = [
    {
        'run': {
            'subs': [
                {
                    'execute': {
                        'help': 'Execute one or more model',
                        'func': run_execute,
                        'parents': [args_input],
                    }
                },
                {
                    'bootstrap': {
                        'help': 'Bootstrap',
                        'func': run_bootstrap,
                        'parents': [args_model_input],
                        'args': [
                            {
                                'name': '--samples',
                                'type': int,
                                'help': 'Number of bootstrap datasets',
                            },
                        ],
                    }
                },
                {
                    'modelsearch': {
                        'help': 'Search for best model',
                        'func': run_modelsearch,
                        'parents': [args_model_input],
                        'args': [
                            {
                                'name': 'algorithm',
                                'type': str,
                                'help': 'Name of algorithm to use',
                            },
                            {
                                'name': 'funcs',
                                'type': semicolon_list,
                                'help': 'List of function calls from modeling',
                            },
                            {
                                'name': '--rankfunc',
                                'type': str,
                                'help': 'Name of function to use for ranking '
                                'candidates (default ofv).',
                                'default': 'ofv',
                            },
                        ],
                    }
                },
            ],
            'help': 'Run a method',
            'title': 'Pharmpy commands for running tools',
            'metavar': 'METHOD',
        }
    },
    {
        'data': {
            'subs': [
                {
                    'write': {
                        'help': 'Write dataset',
                        'parents': [args_model_or_data_input, args_output],
                        'func': data_write,
                        'description': 'Write a dataset from model or csv file as the model sees '
                        'it. For NM-TRAN models this means to filter all IGNORE and '
                        'ACCEPT statements in $DATA and to convert the dataset to '
                        'csv format.',
                    }
                },
                {
                    'print': {
                        'help': 'Print dataset to stdout',
                        'parents': [args_model_or_data_input],
                        'func': data_print,
                        'description': 'Print whole dataset or selected columns from model or csv '
                        'file via a pager to stdout. For NM-TRAN models the dataset will '
                        'be filtered first.',
                        'args': [
                            {
                                'name': '--columns',
                                'metavar': 'COLUMNS',
                                'type': comma_list,
                                'help': 'Select specific columns (default is all)',
                            }
                        ],
                    }
                },
                {
                    'filter': {
                        'help': 'Filter rows of dataset',
                        'description': 'Filter rows of dataset via expressions. '
                        'All rows matching all expressions will be kept.abs '
                        'A new model with the filtered dataset connected is created.',
                        'parents': [args_model_or_data_input, args_output],
                        'func': data_filter,
                        'args': [{'name': 'expressions', 'nargs': argparse.REMAINDER}],
                    }
                },
                {
                    'append': {
                        'help': 'Append column to dataset',
                        'description': 'Append a column to dataset given an assignment expression.'
                        'The expression can contain already present columns of the dataset.',
                        'parents': [args_model_or_data_input, args_output],
                        'func': data_append,
                        'args': [{'name': 'expression'}],
                    }
                },
                {
                    'resample': {
                        'help': 'Resample dataset',
                        'description': 'Bootstrap resample datasets'
                        'Multiple new models and datasets will be created',
                        'parents': [args_model_or_data_input, args_random],
                        'func': data_resample,
                        'args': [
                            {
                                'name': '--group',
                                'metavar': 'COLUMN',
                                'type': str,
                                'default': 'ID',
                                'help': 'Column to use for grouping (default is ID)',
                            },
                            {
                                'name': '--resamples',
                                'metavar': 'NUMBER',
                                'type': int,
                                'default': 1,
                                'help': 'Number of resampled datasets (default 1)',
                            },
                            {
                                'name': '--stratify',
                                'metavar': 'COLUMN',
                                'type': str,
                                'help': 'Column to use for stratification',
                            },
                            {
                                'name': '--replace',
                                'action': 'store_true',
                                'help': 'Sample with replacement (default is without)',
                            },
                            {
                                'name': '--sample_size',
                                'metavar': 'NUMBER',
                                'type': int,
                                'default': None,
                                'help': 'Number of groups to sample for each resample',
                            },
                        ],
                    }
                },
                {
                    'anonymize': {
                        'help': 'Anonymize dataset by renumbering a specific data column',
                        'description': 'Anonymize dataset by renumbering a specific data column',
                        'parents': [args_output],
                        'func': data_anonymize,
                        'args': [
                            {
                                'name': 'dataset',
                                'metavar': 'FILE',
                                'type': str,
                                'help': 'A csv file dataset',
                            },
                            {
                                'name': '--group',
                                'metavar': 'COLUMN',
                                'type': str,
                                'default': 'ID',
                                'help': 'column to anonymize (default ID)',
                            },
                        ],
                    }
                },
            ],
            'help': 'Data manipulations',
            'title': 'Pharmpy data commands',
            'metavar': 'ACTION',
        }
    },
    {'info': {'help': 'Show Pharmpy information', 'title': 'Pharmpy information', 'func': info}},
    {
        'model': {
            'subs': [
                {
                    'print': {
                        'help': 'Format and print model overview',
                        'description': 'Print an overview of a model.',
                        'func': model_print,
                        'parents': [args_input],
                        'args': [
                            {
                                'name': '--explicit-odes',
                                'action': 'store_true',
                                'help': 'Print the ODE system explicitly instead of as a '
                                'compartmental graph',
                            }
                        ],
                    }
                },
                {
                    'sample': {
                        'help': 'Sampling of parameter inits using uncertainty given by covariance '
                        'matrix',
                        'description': 'Sample parameter initial estimates using uncertainty given'
                        'by covariance matrix.',
                        'func': model_sample,
                        'parents': [args_model_input, args_random],
                        'args': [
                            {
                                'name': '--samples',
                                'metavar': 'NUMBER',
                                'type': int,
                                'default': 1,
                                'help': 'Number of sampled models',
                            }
                        ],
                    }
                },
                {
                    'add_cov_effect': {
                        'help': 'Adds covariate effect',
                        'description': 'Add covariate effect to a model parameter',
                        'func': add_covariate_effect,
                        'parents': [args_model_input, args_output],
                        'args': [
                            {'name': 'param', 'type': str, 'help': 'Individual parameter'},
                            {'name': 'covariate', 'type': str, 'help': 'Covariate'},
                            {'name': 'effect', 'type': str, 'help': 'Type of covariate effect'},
                            {
                                'name': '--operation',
                                'type': str,
                                'default': '*',
                                'help': 'Whether effect should be added or multiplied',
                            },
                        ],
                    }
                },
                {
                    'absorption_rate': {
                        'help': 'Set absorption rate for a PK model',
                        'description': 'Change absorption rate of a PK model to either bolus, 0th '
                        'order, 1th order or sequential 0-order 1-order.',
                        'func': model_absorption_rate,
                        'parents': [args_model_input, args_output],
                        'args': [
                            {
                                'name': 'rate',
                                'choices': ['bolus', 'ZO', 'FO', 'seq-ZO-FO'],
                                'type': str,
                                'help': 'Absorption rate',
                            },
                        ],
                    }
                },
                {
                    'elimination_rate': {
                        'help': 'Set elimination rate for a PK model',
                        'description': 'Change elimination rate of a PK model to either 0th '
                        'order, 1th order, Michaelis-Menten or combined 1-order-Michaelis-Menten.',
                        'func': model_elimination_rate,
                        'parents': [args_model_input, args_output],
                        'args': [
                            {
                                'name': 'rate',
                                'choices': ['ZO', 'FO', 'MM', 'comb-FO-MM'],
                                'type': str,
                                'help': 'Eliminiation rate',
                            },
                        ],
                    }
                },
                {
                    'transit_compartments': {
                        'help': 'Set the number of transit compartments for a PK model',
                        'description': 'Set the number of transit compartments of a PK model',
                        'func': model_transit_compartments,
                        'parents': [args_model_input, args_output],
                        'args': [
                            {
                                'name': 'n',
                                'type': int,
                                'help': 'Number of compartments',
                            },
                        ],
                    }
                },
                {
                    'peripheral_compartments': {
                        'help': 'Set the number of peripheral compartments for a PK model',
                        'description': 'Set the number of peripheral compartments of a PK model',
                        'func': model_peripheral_compartments,
                        'parents': [args_model_input, args_output],
                        'args': [
                            {
                                'name': 'n',
                                'type': int,
                                'help': 'Number of compartments',
                            },
                        ],
                    }
                },
                {
                    'error': {
                        'help': 'Set error model',
                        'description': 'Set the error model to either additive, proportional or '
                        'combined',
                        'func': model_error,
                        'parents': [args_model_input, args_output],
                        'args': [
                            {
                                'name': 'error_type',
                                'choices': ['none', 'additive', 'propotional', 'combined'],
                                'type': str,
                                'help': 'Type of error model',
                            },
                        ],
                    }
                },
                {
                    'boxcox': {
                        'help': 'Applies boxcox transformation',
                        'description': 'Apply boxcox transformation on selected/all etas',
                        'func': boxcox,
                        'parents': [args_model_input, args_output],
                        'args': [
                            {
                                'name': '--etas',
                                'type': str,
                                'default': None,
                                'help': 'List of etas, mark group of etas in single quote '
                                'separated by spaces',
                            },
                        ],
                    }
                },
                {
                    'tdist': {
                        'help': 'Applies t-distribution transformation',
                        'description': 'Apply t-distribution transformation on selected/all etas',
                        'func': tdist,
                        'parents': [args_model_input, args_output],
                        'args': [
                            {
                                'name': '--etas',
                                'type': str,
                                'default': None,
                                'help': 'List of etas, mark group of etas in single quote '
                                'separated by spaces',
                            },
                        ],
                    }
                },
                {
                    'john_draper': {
                        'help': 'Applies John Draper transformation',
                        'description': 'Apply John Draper transformation on selected/all etas',
                        'func': john_draper,
                        'parents': [args_model_input, args_output],
                        'args': [
                            {
                                'name': '--etas',
                                'type': str,
                                'default': None,
                                'help': 'List of etas, mark group of etas in single quote '
                                'separated by spaces',
                            },
                        ],
                    }
                },
                {
                    'add_iiv': {
                        'help': 'Adds IIVs',
                        'description': 'Add IIVs to a model parameter',
                        'func': add_iiv,
                        'parents': [args_model_input, args_output],
                        'args': [
                            {'name': 'param', 'type': str, 'help': 'Individual parameter'},
                            {'name': 'expression', 'type': str, 'help': 'Expression for added eta'},
                            {
                                'name': '--operation',
                                'type': str,
                                'default': '*',
                                'help': 'Whether effect should be added or multiplied',
                            },
                            {
                                'name': '--eta_name',
                                'type': str,
                                'default': None,
                                'help': 'Optional custom name of new eta',
                            },
                        ],
                    }
                },
                {
                    'add_iov': {
                        'help': 'Adds IOVs',
                        'description': 'Add IOVs to a random variable or a specified model '
                        'parameter',
                        'func': add_iov,
                        'parents': [args_model_input, args_output],
                        'args': [
                            {
                                'name': 'occ',
                                'type': str,
                                'help': 'Name of occasion column',
                            },
                            {
                                'name': '--etas',
                                'type': str,
                                'default': None,
                                'help': 'List of etas or parameters, mark group of etas in single '
                                'quote separated by spaces',
                            },
                            {
                                'name': '--eta_names',
                                'type': str,
                                'default': None,
                                'help': 'Optional custom names of new etas. Must be equal to the '
                                'number of input etas times the number of categories for '
                                'occasion.',
                            },
                        ],
                    }
                },
                {
                    'create_joint_distribution': {
                        'help': 'Combine etas into joint distributions',
                        'description': 'Creates full or partial block structures of etas (IIV)',
                        'func': create_joint_distribution,
                        'parents': [args_model_input, args_output],
                        'args': [
                            {
                                'name': '--etas',
                                'type': str,
                                'default': None,
                                'help': 'List of etas, mark group of etas in single quote '
                                'separated by spaces. To combine some etas provide list, '
                                'for all etas no arguments are needed',
                            },
                        ],
                    }
                },
                {
                    'iiv_on_ruv': {
                        'help': 'Applies IIV to RUVs',
                        'description': 'Applies IIV to RUVs by multiplying epsilons with '
                        'exponential etas',
                        'func': iiv_on_ruv,
                        'parents': [args_model_input, args_output],
                        'args': [
                            {
                                'name': '--eps',
                                'type': str,
                                'default': None,
                                'help': 'List of epsilons, mark group of epsilons in single quote '
                                'separated by spaces. To apply to all epsilons, omit this '
                                'argument.',
                            },
                            {
                                'name': '--same_eta',
                                'type': bool,
                                'default': True,
                                'help': 'whether all RUVs from input should use the same new ETA '
                                'or if one ETA should be created for each RUV.',
                            },
                            {
                                'name': '--eta_names',
                                'type': str,
                                'default': None,
                                'help': 'Optional custom names of new etas. Must be equal to the '
                                'number of input etas times the number of categories for '
                                'occasion.',
                            },
                        ],
                    }
                },
                {
                    'remove_iiv': {
                        'help': 'Removes IIVs',
                        'description': 'Removes all IIV omegas given a list with eta names '
                        'and/or parameter names',
                        'func': remove_iiv,
                        'parents': [args_model_input, args_output],
                        'args': [
                            {
                                'name': '--to_remove',
                                'type': str,
                                'default': None,
                                'help': 'List of etas and/or parameter name to remove. To remove '
                                'all etas, omit this argument.',
                            },
                        ],
                    }
                },
                {
                    'remove_iov': {
                        'help': 'Removes IOVs',
                        'description': 'Removes all IOV omegas',
                        'func': remove_iov,
                        'parents': [args_model_input, args_output],
                    }
                },
                {
                    'power_on_ruv': {
                        'help': 'Applies power effect to RUVs.',
                        'description': 'Applies a power effect to provided epsilons.',
                        'func': power_on_ruv,
                        'parents': [args_model_input, args_output],
                        'args': [
                            {
                                'name': '--eps',
                                'type': str,
                                'default': None,
                                'help': 'List of epsilons, mark group of epsilons in single quote '
                                'separated by spaces. To apply to all epsilons, omit this '
                                'argument.',
                            },
                        ],
                    }
                },
                {
                    'update_inits': {
                        'help': 'Update inits using modelfit results.',
                        'description': 'Update inits using modelfit results, can be forced.',
                        'func': update_inits,
                        'parents': [args_model_input, args_output],
                        'args': [
                            {
                                'name': '--force_update',
                                'type': bool,
                                'default': False,
                                'help': 'Force update.',
                            },
                        ],
                    }
                },
            ],
            'help': 'Model manipulations',
            'title': 'Pharmpy model commands',
            'metavar': 'ACTION',
        }
    },
    {
        'results': {
            'subs': [
                {
                    'bootstrap': {
                        'help': 'Generate bootstrap results',
                        'description': 'Generate results from a PsN bootstrap run',
                        'func': results_bootstrap,
                        'args': [
                            {
                                'name': 'psn_dir',
                                'metavar': 'PsN directory',
                                'type': pathlib.Path,
                                'help': 'Path to PsN bootstrap run directory',
                            }
                        ],
                    }
                },
                {
                    'cdd': {
                        'help': 'Generate cdd results',
                        'description': 'Generate results from a PsN cdd run',
                        'func': results_cdd,
                        'args': [
                            {
                                'name': 'psn_dir',
                                'metavar': 'PsN directory',
                                'type': pathlib.Path,
                                'help': 'Path to PsN cdd run directory',
                            }
                        ],
                    }
                },
                {
                    'frem': {
                        'help': 'Generate FREM results',
                        'description': 'Generate results from a PsN frem run',
                        'func': results_frem,
                        'args': [
                            {
                                'name': 'psn_dir',
                                'metavar': 'PsN directory',
                                'type': pathlib.Path,
                                'help': 'Path to PsN frem run directory',
                            },
                            {
                                'name': '--method',
                                'choices': ['cov_sampling', 'bipp'],
                                'default': 'cov_sampling',
                                'help': 'Method to use for uncertainty of covariate effects',
                            },
                            {
                                'name': '--force_posdef_covmatrix',
                                'action': 'store_true',
                                'help': 'Should covariance matrix be forced to become positive '
                                'definite',
                            },
                            {
                                'name': '--force_posdef_samples',
                                'type': int,
                                'default': 500,
                                'help': 'Number of sampling tries to do before starting to force '
                                'posdef',
                            },
                        ],
                    }
                },
                {
                    'linearize': {
                        'help': 'Generate linearize results',
                        'description': 'Generate results from a PsN linearize run',
                        'func': results_linearize,
                        'args': [
                            {
                                'name': 'psn_dir',
                                'metavar': 'PsN directory',
                                'type': pathlib.Path,
                                'help': 'Path to PsN linearize run directory',
                            }
                        ],
                    }
                },
                {
                    'ofv': {
                        'help': 'Extract OFVs from model runs',
                        'description': 'Extract OFVs from one or more model runs',
                        'parents': [args_input],
                        'func': results_ofv,
                    }
                },
                {
                    'print': {
                        'help': 'Print results',
                        'description': 'Print results from PsN run to stdout',
                        'func': results_print,
                        'args': [
                            {
                                'name': 'dir',
                                'metavar': 'file or directory',
                                'type': pathlib.Path,
                                'help': 'Path to directory containing results.json '
                                'or directly to json results file',
                            }
                        ],
                    }
                },
                {
                    'qa': {
                        'help': 'Generate qa results',
                        'description': 'Generate results from a PsN qa run',
                        'func': results_qa,
                        'args': [
                            {
                                'name': 'psn_dir',
                                'metavar': 'PsN directory',
                                'type': pathlib.Path,
                                'help': 'Path to PsN qa run directory',
                            }
                        ],
                    }
                },
                {
                    'report': {
                        'help': 'Generate results report',
                        'description': 'Generate results report for PsN run (currently only frem)',
                        'func': results_report,
                        'args': [
                            {
                                'name': 'psn_dir',
                                'metavar': 'PsN directory',
                                'type': pathlib.Path,
                                'help': 'Path to PsN run directory',
                            }
                        ],
                    }
                },
                {
                    'resmod': {
                        'help': 'Generate resmod results',
                        'description': 'Generate results from a PsN resmod run',
                        'func': results_resmod,
                        'args': [
                            {
                                'name': 'psn_dir',
                                'metavar': 'PsN directory',
                                'type': pathlib.Path,
                                'help': 'Path to PsN resmod run directory',
                            }
                        ],
                    }
                },
                {
                    'scm': {
                        'help': 'Generate scm results',
                        'description': 'Generate results from a PsN scm run',
                        'func': results_scm,
                        'args': [
                            {
                                'name': 'psn_dir',
                                'metavar': 'PsN directory',
                                'type': pathlib.Path,
                                'help': 'Path to PsN scm run directory',
                            }
                        ],
                    }
                },
                {
                    'simeval': {
                        'help': 'Generate simeval results',
                        'description': 'Generate results from a PsN simeval run',
                        'func': results_simeval,
                        'args': [
                            {
                                'name': 'psn_dir',
                                'metavar': 'PsN directory',
                                'type': pathlib.Path,
                                'help': 'Path to PsN simeval run directory',
                            }
                        ],
                    }
                },
                {
                    'summary': {
                        'help': 'Modelfit summary',
                        'description': 'Print a summary of a model estimates to stdout.',
                        'parents': [args_input],
                        'func': results_summary,
                    }
                },
            ],
            'help': 'Result extraction and generation',
            'title': 'Pharmpy result generation commands',
            'metavar': 'ACTION',
        }
    },
]


def generate_parsers(parsers):
    for command in parser_definition:
        ((cmd_name, cmd_dict),) = command.items()
        cmd_parser = parsers.add_parser(
            cmd_name, allow_abbrev=True, help=cmd_dict['help'], formatter_class=formatter
        )
        if 'subs' in cmd_dict:
            subs = cmd_parser.add_subparsers(title=cmd_dict['title'], metavar=cmd_dict['metavar'])
            for sub_command in cmd_dict['subs']:
                ((sub_name, sub_dict),) = sub_command.items()
                args = sub_dict.pop('args', [])
                func = sub_dict.pop('func')

                sub_parser = subs.add_parser(sub_name, **sub_dict, formatter_class=formatter)
                for arg in args:
                    name = arg.pop('name')
                    sub_parser.add_argument(name, **arg)
                sub_parser.set_defaults(func=func)


parser = argparse.ArgumentParser(
    prog='pharmpy',
    description=dedent(
        """
    Welcome to the command line interface of Pharmpy!

    Functionality is split into various subcommands
        - try --help after a COMMAND
        - all keyword arguments can be abbreviated if unique


    """
    ).strip(),
    epilog=dedent(
        """
        Examples:
            # Create 100 bootstrap datasets
            pharmpy data resample pheno_real.mod --resamples=100 --replace

            # prettyprint model
            pharmpy model print pheno_real.mod

            # version/install information
            pharmpy info
    """
    ).strip(),
    formatter_class=formatter,
    allow_abbrev=True,
)
parser.add_argument('--version', action='version', version='0.37.0')

# subcommand parsers
subparsers = parser.add_subparsers(title='Pharmpy commands', metavar='COMMAND')
generate_parsers(subparsers)


# -- entry point of CLI (pharmpy) ---------------------------------------------------------


def main(args):
    """Entry point of ``pharmpy`` CLI util and ``python3 -m pharmpy`` (via ``__main__.py``)."""
    # parse
    args = parser.parse_args(args)
    # dispatch subcommand
    if 'func' in args:
        args.func(args)
    else:
        parser.print_usage()



"""
:meta private:
"""

from operator import add, mul

import sympy
from sympy import Eq, Piecewise

from pharmpy.modeling.help_functions import _format_input_list, _format_options, _get_etas
from pharmpy.parameter import Parameter
from pharmpy.random_variables import RandomVariable
from pharmpy.statements import Assignment, ModelStatements
from pharmpy.symbols import symbol as S


def add_iiv(model, list_of_parameters, expression, operation='*', eta_names=None):
    """Adds IIVs to :class:`pharmpy.model`.

    Effects that currently have templates are:

    - Additive (*add*)
    - Proportional (*prop*)
    - Exponential (*exp*)
    - Logit (*log*)

    For all except exponential the operation input is not needed. Otherwise user specified
    input is supported. Initial estimates for new etas are 0.09.

    Parameters
    ----------
    model : Model
        Pharmpy model to add new IIVs to.
    list_of_parameters : str, list
        Name/names of parameter to add new IIVs to.
    expression : str, list
        Effect/effects on eta. Either abbreviated (see above) or custom.
    operation : str, list, optional
        Whether the new IIV should be added or multiplied (default).
    eta_names : str, list, optional
        Custom name/names of new eta

    Return
    ------
    Model
        Reference to the same model

    Example
    -------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> remove_iiv(model, "CL") # doctest: +ELLIPSIS
    <...>
    >>> add_iiv(model, "CL", "add")  # doctest: +ELLIPSIS
    <...>
    >>> model.statements.find_assignment("CL")
    CL := ETA_CL + TVCL

    See also
    --------
    add_iov
    remove_iiv
    remove_iov

    """
    rvs, pset, sset = model.random_variables, model.parameters, model.statements

    list_of_parameters = _format_input_list(list_of_parameters)
    list_of_options = _format_options([expression, operation, eta_names], len(list_of_parameters))
    expression, operation, eta_names = list_of_options

    if all(eta_name is None for eta_name in eta_names):
        eta_names = None

    if not all(len(opt) == len(list_of_parameters) for opt in list_of_options if opt):
        raise ValueError(
            'The number of provided expressions and operations must either be equal '
            'to the number of parameters or equal to 1'
        )

    for i in range(len(list_of_parameters)):
        omega = S(f'IIV_{list_of_parameters[i]}')
        if not eta_names:
            eta_name = f'ETA_{list_of_parameters[i]}'
        else:
            eta_name = eta_names[i]

        eta = RandomVariable.normal(eta_name, 'iiv', 0, omega)

        rvs.append(eta)
        pset.append(Parameter(str(omega), init=0.09))

        statement = sset.find_assignment(list_of_parameters[i])

        eta_addition = _create_template(expression[i], operation[i])
        eta_addition.apply(statement.expression, eta.name)

        statement.expression = eta_addition.template

    model.random_variables = rvs
    model.parameters = pset
    model.statements = sset

    return model


def add_iov(model, occ, list_of_parameters=None, eta_names=None):
    """Adds IOVs to :class:`pharmpy.model`.

    Initial estimate of new IOVs are 10% of the IIV eta it is based on.

    Parameters
    ----------
    model : Model
        Pharmpy model to add new IOVs to.
    occ : str
        Name of occasion column.
    list_of_parameters : str, list
        List of names of parameters and random variables. Accepts random variable names, parameter
        names, or a mix of both.
    eta_names : str, list
        Custom names of new etas. Must be equal to the number of input etas times the number of
        categories for occasion.

    Return
    ------
    Model
        Reference to the same model

    Example
    -------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> add_iov(model, "TIME", "CL")  # doctest: +SKIP
    <...>
    >>> model.statements.find_assignment("CL")  # doctest: +SKIP
    CL := ETA_CL + TVCL

    See also
    --------
    add_iiv
    remove_iiv
    remove_iov

    """
    rvs, pset, sset = model.random_variables, model.parameters, model.statements

    list_of_parameters = _format_input_list(list_of_parameters)
    etas = _get_etas(model, list_of_parameters, include_symbols=True)
    categories = _get_occ_levels(model.dataset, occ)

    if eta_names and len(eta_names) != len(etas) * len(categories):
        raise ValueError(
            f'Number of provided names incorrect, need {len(etas) * len(categories)} names.'
        )
    elif len(categories) == 1:
        raise ValueError(f'Only one value in {occ} column.')

    iovs, etais = ModelStatements(), ModelStatements()
    for i, eta in enumerate(etas, 1):
        omega_name = str(next(iter(eta.sympy_rv.pspace.distribution.free_symbols)))
        omega = S(f'OMEGA_IOV_{i}')  # TODO: better name
        pset.append(Parameter(str(omega), init=pset[omega_name].init * 0.1))

        iov = S(f'IOV_{i}')

        values, conditions = [], []

        for j, cat in enumerate(categories, 1):
            if eta_names:
                eta_name = eta_names[j - 1]
            else:
                eta_name = f'ETA_IOV_{i}{j}'

            eta_new = RandomVariable.normal(eta_name, 'iov', 0, omega)
            rvs.append(eta_new)

            values += [S(eta_new.name)]
            conditions += [Eq(cat, S(occ))]

        expression = Piecewise(*zip(values, conditions))

        iovs.append(Assignment(iov, sympy.sympify(0)))
        iovs.append(Assignment(iov, expression))
        etais.append(Assignment(S(f'ETAI{i}'), eta.symbol + iov))

        sset.subs({eta.name: S(f'ETAI{i}')})

    iovs.extend(etais)
    iovs.extend(sset)

    model.random_variables, model.parameters, model.statements = rvs, pset, iovs

    return model


def _create_template(expression, operation):
    operation_func = _get_operation_func(operation)
    if expression == 'add':
        return EtaAddition.additive()
    elif expression == 'prop':
        return EtaAddition.proportional()
    elif expression == 'exp':
        return EtaAddition.exponential(operation_func)
    elif expression == 'log':
        return EtaAddition.logit()
    else:
        expression = sympy.sympify(f'original {operation} {expression}')
        return EtaAddition(expression)


def _get_operation_func(operation):
    """Gets sympy operation based on string"""
    if operation == '*':
        return mul
    elif operation == '+':
        return add


def _get_occ_levels(df, occ):
    levels = df[occ].unique()
    return _round_categories(levels)


def _round_categories(categories):
    categories_rounded = []
    for c in categories:
        if not isinstance(c, int) or c.is_integer():
            categories_rounded.append(int(c))
        else:
            categories_rounded.append(c)
    categories_rounded.sort()
    return categories_rounded


class EtaAddition:
    """
    Eta addition consisting of expression.

    Attributes
    ----------
    template
        Expression consisting of original statement +/* new eta with effect.

    :meta private:

    """

    def __init__(self, template):
        self.template = template

    def apply(self, original, eta):
        self.template = self.template.subs({'original': original, 'eta_new': eta})

    @classmethod
    def additive(cls):
        template = S('original') + S('eta_new')

        return cls(template)

    @classmethod
    def proportional(cls):
        template = S('original') * S('eta_new')

        return cls(template)

    @classmethod
    def exponential(cls, operation):
        template = operation(S('original'), sympy.exp(S('eta_new')))

        return cls(template)

    @classmethod
    def logit(cls):
        template = S('original') * (sympy.exp(S('eta_new')) / (1 + sympy.exp(S('eta_new'))))

        return cls(template)



from pathlib import Path

import numpy as np
import pandas as pd
import yaml

import pharmpy.random_variables
import pharmpy.symbols
import pharmpy.tools.psn_helpers as psn_helpers
from pharmpy import Model
from pharmpy.results import Results, read_results


class QAResults(Results):
    def __init__(
        self,
        dofv=None,
        fullblock_parameters=None,
        boxcox_parameters=None,
        tdist_parameters=None,
        add_etas_parameters=None,
        iov_parameters=None,
        influential_individuals=None,
        covariate_effects=None,
        univariate_sum=None,
        residual_error=None,
        structural_bias=None,
    ):
        self.dofv = dofv
        self.fullblock_parameters = fullblock_parameters
        self.boxcox_parameters = boxcox_parameters
        self.tdist_parameters = tdist_parameters
        self.add_etas_parameters = add_etas_parameters
        self.iov_parameters = iov_parameters
        self.influential_individuals = influential_individuals
        self.covariate_effects = covariate_effects
        self.univariate_sum = univariate_sum
        self.residual_error = residual_error
        self.structural_bias = structural_bias


def calculate_results(
    original_model,
    base_model,
    fullblock_model=None,
    boxcox_model=None,
    tdist_model=None,
    add_etas_model=None,
    iov_model=None,
    etas_added_to=None,
    frem_results=None,
    cdd_results=None,
    scm_results=None,
    simeval_results=None,
    resmod_idv_results=None,
):
    fullblock_table, fullblock_dofv = calc_fullblock(original_model, fullblock_model)
    boxcox_table, boxcox_dofv = calc_transformed_etas(
        original_model, boxcox_model, 'boxcox', 'lambda'
    )
    tdist_table, tdist_dofv = calc_transformed_etas(original_model, tdist_model, 'tdist', 'df')
    addetas_table, addetas_dofv = calc_add_etas(original_model, add_etas_model, etas_added_to)
    iov_table, iov_dofv = calc_iov(original_model, iov_model)
    frem_dofv = calc_frem_dofv(base_model, fullblock_model, frem_results)
    univariate_sum, scm_table, scm_dofv = calc_scm_dofv(scm_results)
    infinds, cdd_dofv = influential_individuals(cdd_results)
    outs, simeval_dofv = outliers(simeval_results, cdd_results)
    resmodtab, resmod_dofv = resmod(resmod_idv_results)

    dofv_table = pd.concat(
        [
            fullblock_dofv,
            boxcox_dofv,
            tdist_dofv,
            addetas_dofv,
            iov_dofv,
            frem_dofv,
            scm_dofv,
            cdd_dofv,
            simeval_dofv,
            resmod_dofv,
        ]
    )
    dofv_table.set_index(['section', 'run', 'dvid'], inplace=True)
    res = QAResults(
        dofv=dofv_table,
        fullblock_parameters=fullblock_table,
        boxcox_parameters=boxcox_table,
        tdist_parameters=tdist_table,
        add_etas_parameters=addetas_table,
        iov_parameters=iov_table,
        influential_individuals=infinds,
        covariate_effects=scm_table,
        univariate_sum=univariate_sum,
        residual_error=resmodtab,
    )
    return res


def outliers(simeval_res, cdd_res):
    dofv_tab = pd.DataFrame(
        {
            'section': ['outliers'],
            'run': ['1'],
            'dofv': [np.nan],
            'df': [np.nan],
        }
    )
    if simeval_res is None or cdd_res is None:
        return None, dofv_tab
    iofv = simeval_res.iofv_summary
    outliers = list(iofv.loc[iofv['residual_outlier']].index)
    cases = cdd_res.case_results.copy()
    cases['skipped_individuals'] = cases['skipped_individuals'].transform(lambda x: int(x[0]))
    cases.reset_index(inplace=True)
    cases.set_index('skipped_individuals', inplace=True)
    dofv = cases.loc[outliers].delta_ofv
    top_three = dofv.sort_values(ascending=False)[:3]
    dofv_tab = pd.DataFrame(
        {
            'section': ['outliers'] * len(top_three),
            'run': list(top_three.index),
            'dofv': top_three,
            'df': [np.nan] * len(top_three),
        }
    )

    return None, dofv_tab


def influential_individuals(cdd_res):
    dofv_tab = pd.DataFrame(
        {
            'section': ['influential_individuals'],
            'run': ['1'],
            'dofv': [np.nan],
            'df': [np.nan],
        }
    )
    if cdd_res is None:
        return None, dofv_tab
    df = cdd_res.case_results
    df = df.loc[df['delta_ofv'] > 3.84]
    skipped = [e[0] for e in df['skipped_individuals']]
    influentials = pd.DataFrame({'delta_ofv': df['delta_ofv']}, index=skipped)
    influentials.index.name = 'ID'
    top_three = influentials.sort_values(by=['delta_ofv']).iloc[:3]
    dofv_tab = pd.DataFrame(
        {
            'section': ['influential_individuals'] * len(top_three),
            'run': list(top_three.index),
            'dofv': top_three['delta_ofv'],
            'df': [np.nan] * len(top_three),
        }
    )
    return influentials, dofv_tab


def resmod(res):
    dofv_tab = pd.DataFrame(
        {
            'section': ['residual_error_model'],
            'run': [np.nan],
            'dvid': [np.nan],
            'dofv': [np.nan],
            'df': [np.nan],
        }
    )
    if res is None:
        return None, dofv_tab

    df = res.models.copy()
    df = df.droplevel(0)
    df.drop('sum', level='DVID', errors='ignore', inplace=True)
    df['dOFV'] = -df['dOFV']
    df.drop(
        ['idv_varying_RUV', 'idv_varying_combined', 'idv_varying_theta'],
        level='model',
        inplace=True,
    )

    # Select the best idv_varying cut for each DVID
    remaining_timevar = []
    for dvid in df.index.unique(level='DVID'):
        subdf = df.loc[dvid]
        idvvar = subdf.index.str.startswith("idv_varying")
        best_timevar_idx = subdf.loc[idvvar]['dOFV'].idxmax()
        for name in subdf.loc[idvvar].index:
            if name != best_timevar_idx:
                df.drop((dvid, name), inplace=True)
        remaining_timevar.append(best_timevar_idx)
    for ind in remaining_timevar:
        df.rename(index={ind: 'time_varying'}, level='model', inplace=True)

    df = df.groupby(level='DVID', sort=False).apply(
        lambda x: x.sort_values(['dOFV'], ascending=False)
    )
    df = df.droplevel(0)  # FIXME: Why was an extra DVID level added?
    n = df['parameters'].apply(lambda x: len(x))
    df.insert(1, 'additional_parameters', n)
    for dvid in df.index.unique(level='DVID'):
        df.loc[(dvid, 'time_varying'), 'additional_parameters'] = 2

    inds = list(df.index[:2])
    dvid2 = [dvid for dvid, _ in inds]
    run2 = [run for _, run in inds]

    dofv_tab = pd.DataFrame(
        {
            'section': ['residual_error_model'] * 2,
            'run': run2,
            'dvid': dvid2,
            'dofv': list(df['dOFV'].iloc[:2]),
            'df': list(df['additional_parameters'].iloc[:2]),
        }
    )

    return df, dofv_tab


def calc_iov(original_model, iov_model):
    dofv_tab = pd.DataFrame(
        {
            'section': ['parameter_variability'],
            'run': ['iov'],
            'dvid': [np.nan],
            'dofv': [np.nan],
            'df': [np.nan],
        }
    )
    if iov_model is None:
        return None, dofv_tab
    iov_res = iov_model.modelfit_results
    if iov_res is None:
        return None, dofv_tab
    origres = original_model.modelfit_results
    iov_params = iov_model.random_variables.iov.variance_parameters
    iov_sds = [iov_res.parameter_estimates_sdcorr[param] for param in iov_params]
    iiv_params = iov_model.random_variables.iiv.variance_parameters
    iiv_params = [param for param in iiv_params if not original_model.parameters[param].fix]
    new_iiv_sds = [iov_res.parameter_estimates_sdcorr[param] for param in iiv_params]
    old_iiv_sds = [origres.parameter_estimates_sdcorr[param] for param in iiv_params]

    etas = []
    for rvs, dist in original_model.random_variables.iiv.distributions():
        if not set(iiv_params).isdisjoint({s.name for s in dist.free_symbols}):
            etas.extend(rvs)
    etas = [eta.name for eta in etas]

    table = pd.DataFrame(
        {'new_iiv_sd': new_iiv_sds, 'orig_iiv_sd': old_iiv_sds, 'iov_sd': iov_sds}, index=etas
    )

    dofv = origres.ofv - iov_res.ofv
    dofv_tab = pd.DataFrame(
        {
            'section': ['parameter_variability'],
            'run': ['iov'],
            'dvid': [np.nan],
            'dofv': [dofv],
            'df': [len(iov_params)],
        }
    )
    return table, dofv_tab


def calc_add_etas(original_model, add_etas_model, etas_added_to):
    dofv_tab = pd.DataFrame(
        {
            'section': ['parameter_variability'],
            'run': ['add_etas'],
            'dvid': [np.nan],
            'dofv': [np.nan],
            'df': [np.nan],
        }
    )
    if add_etas_model is None:
        return None, dofv_tab
    add_etas_res = add_etas_model.modelfit_results
    if add_etas_res is None:
        return None, dofv_tab
    origres = original_model.modelfit_results
    original_etas = [rv.name for rv in original_model.random_variables.etas]
    all_etas = original_etas + etas_added_to
    added = [True] * len(original_etas) + [False] * len(etas_added_to)
    params = add_etas_model.random_variables.etas.variance_parameters
    params = [pharmpy.symbols.symbol(p) for p in params]
    orig_params = original_model.random_variables.etas.variance_parameters
    orig_params = [pharmpy.symbols.symbol(p) for p in orig_params]
    add_etas_sds = [add_etas_res.parameter_estimates_sdcorr[p.name] for p in params]
    orig_sds = [origres.parameter_estimates_sdcorr[p.name] for p in orig_params]
    orig_sds += [np.nan] * len(etas_added_to)
    table = pd.DataFrame(
        {'added': added, 'new_sd': add_etas_sds, 'orig_sd': orig_sds}, index=all_etas
    )
    dofv = origres.ofv - add_etas_res.ofv
    dofv_tab = pd.DataFrame(
        {
            'section': ['parameter_variability'],
            'run': ['add_etas'],
            'dvid': [np.nan],
            'dofv': [dofv],
            'df': [len(etas_added_to)],
        }
    )
    return table, dofv_tab


def calc_scm_dofv(scm_results):
    dofv_tab = pd.DataFrame(
        {
            'section': ['covariates'],
            'run': ['scm'],
            'dvid': [np.nan],
            'dofv': [np.nan],
            'df': [np.nan],
        }
    )
    if scm_results is None:
        return None, None, dofv_tab
    table = scm_results.steps.copy()
    table.index = table.index.droplevel(3)
    table.index = table.index.droplevel(0)
    univariate_sum = table['ofv_drop'].sum()
    top = table.sort_values(by=['ofv_drop']).iloc[-1:]
    table['coeff'] = [list(coveff.values())[0] for coveff in table['covariate_effects']]
    table = table[['ofv_drop', 'coeff']]
    table.columns = ['dofv', 'coeff']
    table.rename(mapper=lambda name: f'ETA({name[2:]})', level=0, inplace=True)
    dofv_tab = pd.DataFrame(
        {
            'section': ['covariates'],
            'run': top['model'].values,
            'dvid': [np.nan],
            'dofv': top['ofv_drop'].values,
            'df': top['delta_df'].values,
        }
    )
    return univariate_sum, table, dofv_tab


def calc_frem_dofv(base_model, fullblock_model, frem_results):
    """Calculate the dOFV for the frem model"""
    dofv_tab = pd.DataFrame(
        {
            'section': ['covariates'],
            'run': ['frem'],
            'dvid': [np.nan],
            'dofv': [np.nan],
            'df': [np.nan],
        }
    )
    if base_model is None or frem_results is None:
        return dofv_tab
    baseres = base_model.modelfit_results
    if baseres is not None:
        base_ofv = baseres.ofv
    else:
        return dofv_tab
    if fullblock_model is not None:
        fullres = fullblock_model.modelfit_results
        if fullres is not None:
            full_ofv = fullres.ofv
        else:
            return dofv_tab
    else:
        full_ofv = 0

    model2_ofv = frem_results.ofv['ofv']['model_2']
    model4_ofv = frem_results.ofv['ofv']['model_4']

    dofv = model2_ofv - model4_ofv - (base_ofv - full_ofv)
    npar = len(frem_results.covariate_effects.index.get_level_values('parameter').unique())
    ncov = len(frem_results.covariate_effects.index.get_level_values('covariate').unique())
    dofv_tab = pd.DataFrame(
        {
            'section': ['covariates'],
            'run': ['frem'],
            'dvid': [np.nan],
            'dofv': [dofv],
            'df': [npar * ncov],
        }
    )
    return dofv_tab


def calc_transformed_etas(original_model, new_model, transform_name, parameter_name):
    """Retrieve new and old parameters of boxcox"""
    dofv_tab = pd.DataFrame(
        {
            'section': ['parameter_variability'],
            'run': [transform_name],
            'dvid': [np.nan],
            'dofv': [np.nan],
            'df': [np.nan],
        }
    )
    if new_model is None:
        return None, dofv_tab
    origres = original_model.modelfit_results
    newres = new_model.modelfit_results
    if newres is None:
        return None, dofv_tab
    params = new_model.random_variables.etas.variance_parameters
    params = [pharmpy.symbols.symbol(p) for p in params]
    boxcox_sds = [newres.parameter_estimates_sdcorr[p.name] for p in params]
    orig_sds = [origres.parameter_estimates_sdcorr[p.name] for p in params]
    thetas = newres.parameter_estimates_sdcorr[0 : len(params)]
    eta_names = [rv.name for rv in new_model.random_variables.etas]

    table = pd.DataFrame(
        {parameter_name: thetas.values, 'new_sd': boxcox_sds, 'old_sd': orig_sds}, index=eta_names
    )

    dofv = origres.ofv - newres.ofv
    dofv_tab = pd.DataFrame(
        {
            'section': ['parameter_variability'],
            'run': [transform_name],
            'dofv': [dofv],
            'df': [len(eta_names)],
        }
    )
    return table, dofv_tab


def calc_fullblock(original_model, fullblock_model):
    """Retrieve new and old parameters of full block"""
    dofv_tab = pd.DataFrame(
        {
            'section': ['parameter_variability'],
            'run': ['fullblock'],
            'dvid': [np.nan],
            'dofv': [np.nan],
            'df': [np.nan],
        }
    )
    if fullblock_model is None:
        return None, dofv_tab
    origres = original_model.modelfit_results
    fullres = fullblock_model.modelfit_results
    if fullres is None:
        return None, dofv_tab
    _, dist = fullblock_model.random_variables.iiv.distributions()[0]
    fullblock_parameters = [str(symb) for symb in dist.free_symbols]
    new_params = (
        fullres.parameter_estimates_sdcorr[fullblock_parameters]
        .reindex(index=fullres.parameter_estimates_sdcorr.index)
        .dropna()
    )
    old_params = origres.parameter_estimates_sdcorr
    table = pd.DataFrame({'new': new_params, 'old': old_params}).reindex(index=new_params.index)

    degrees = table['old'].isna().sum()
    dofv = origres.ofv - fullres.ofv
    dofv_tab = pd.DataFrame(
        {
            'section': ['parameter_variability'],
            'run': ['fullblock'],
            'dvid': [np.nan],
            'dofv': [dofv],
            'df': [degrees],
        }
    )
    return table, dofv_tab


def read_results_summary(path):
    summary_path = path / 'results_summary.yaml'
    if not summary_path.is_file():
        return
    with open(summary_path, 'r') as fh:
        summary = yaml.safe_load(fh)
    dfs = []
    for idvtab in summary['structural']:
        idv = idvtab['idv']
        dvid = idvtab['dvid']
        if dvid == 'NA':
            dvid = 1
        cwres = [float(num) for num in idvtab['second_table']['CWRES']]
        cpred = [int(num) for num in idvtab['second_table']['%CPRED']]
        binmin = idvtab['table']['bin_min']
        binmax = idvtab['table']['bin_max']
        df = pd.DataFrame(
            {
                'idv': idv,
                'dvid': dvid,
                'binid': list(range(1, len(cpred) + 1)),
                'binmin': binmin,
                'binmax': binmax,
                'cwres': cwres,
                'cpred': cpred,
            }
        )
        dfs.append(df)
    bias = pd.concat(dfs)
    bias.set_index(['idv', 'dvid', 'binid'], inplace=True)
    return bias


def psn_qa_results(path):
    """Create qa results from a PsN qa run

    :param path: Path to PsN qa run directory
    :return: A :class:`QAResults` object
    """
    path = Path(path)

    original_model = Model(path / 'linearize_run' / 'scm_dir1' / 'derivatives.mod')
    base_path = list(path.glob('*_linbase.mod'))[0]
    base_model = Model(base_path)
    fullblock_path = path / 'modelfit_run' / 'fullblock.mod'
    if fullblock_path.is_file():
        fullblock_model = Model(fullblock_path)
    else:
        fullblock_model = None
    boxcox_path = path / 'modelfit_run' / 'boxcox.mod'
    if boxcox_path.is_file():
        boxcox_model = Model(boxcox_path)
    else:
        boxcox_model = None
    tdist_path = path / 'modelfit_run' / 'tdist.mod'
    if tdist_path.is_file():
        tdist_model = Model(tdist_path)
    else:
        tdist_model = None
    addetas_path = path / 'add_etas_run' / 'add_etas_linbase.mod'
    if addetas_path.is_file():
        addetas_model = Model(addetas_path)
    else:
        addetas_model = None
    iov_path = path / 'modelfit_run' / 'iov.mod'
    if iov_path.is_file():
        iov_model = Model(iov_path)
    else:
        iov_model = None

    frem_path = path / 'frem_run' / 'results.json'
    if frem_path.is_file():
        frem_res = read_results(frem_path)
    else:
        frem_res = None

    cdd_path = path / 'cdd_run' / 'results.json'
    if cdd_path.is_file():
        cdd_res = read_results(cdd_path)
    else:
        cdd_res = None

    scm_path = path / 'scm_run' / 'results.json'
    if scm_path.is_file():
        scm_res = read_results(scm_path)
    else:
        scm_res = None

    simeval_path = path / 'simeval_run' / 'results.json'
    if simeval_path.is_file():
        simeval_res = read_results(simeval_path)
    else:
        simeval_res = None

    args = psn_helpers.options_from_command(psn_helpers.psn_command(path))
    if 'add_etas' not in args:
        etas_added_to = None
    else:
        etas_added_to = args['add_etas'].split(',')

    idv = args.get('resmod_idv', 'TIME')
    resmod_idv_path = path / f'resmod_{idv}' / 'results.json'
    if resmod_idv_path.is_file():
        resmod_idv_res = read_results(resmod_idv_path)
    else:
        resmod_idv_res = None

    res = calculate_results(
        original_model,
        base_model,
        fullblock_model=fullblock_model,
        boxcox_model=boxcox_model,
        tdist_model=tdist_model,
        add_etas_model=addetas_model,
        iov_model=iov_model,
        etas_added_to=etas_added_to,
        frem_results=frem_res,
        cdd_results=cdd_res,
        scm_results=scm_res,
        simeval_results=simeval_res,
        resmod_idv_results=resmod_idv_res,
    )

    bias = read_results_summary(path)
    res.structural_bias = bias

    return res



import pytest
from sympy import exp

from pharmpy.modeling.eta_transformations import EtaTransformation
from pharmpy.symbols import symbol as S


@pytest.mark.parametrize(
    'eta_trans,symbol,expression',
    [
        (
            EtaTransformation.boxcox(2),
            S('ETAB(2)'),
            ((exp(S('ETA(2)')) ** S('BOXCOX2') - 1) / S('BOXCOX2')),
        ),
    ],
)
def test_apply(eta_trans, symbol, expression):
    etas = {'eta1': S('ETA(1)'), 'eta2': S('ETA(2)'), 'etab1': S('ETAB(1)'), 'etab2': S('ETAB(2)')}

    thetas = {'theta1': 'BOXCOX1', 'theta2': 'BOXCOX2'}

    eta_trans.apply(etas, thetas)
    assert eta_trans.assignments[1].symbol == symbol
    assert eta_trans.assignments[1].expression == expression



"""
:meta private:
"""


def update_inits(model, force_individual_estimates=False):
    """Update initial parameter estimate for a model

    Updates initial estimates of population parameters for a model from
    its modelfit_results. If the model has used initial estimates for
    individual estimates these will also be updated. If initial estimates

    Parameters
    ----------
    model : Model
        Pharmpy model to update initial estimates
    force_individual_estimates : bool
        Update initial individual estimates even if model din't use them previously.

    Returns
    -------
    Model
        Reference to the same model

    Example
    -------
    >>> from pharmpy.modeling import load_example_model, update_inits
    >>> model = load_example_model("pheno")
    >>> model.parameters.inits  # doctest:+ELLIPSIS
    {'THETA(1)': 0.00469307, 'THETA(2)': 1.00916, 'THETA(3)': 0.1, 'OMEGA(1,1)': 0.0309626...}
    >>> update_inits(model)  # doctest:+ELLIPSIS
    <...>
    >>> model.parameters.inits  # doctest:+ELLIPSIS
    {'THETA(1)': 0.00469555, 'THETA(2)': 0.984258, 'THETA(3)': 0.15892, 'OMEGA(1,1)': 0.0293508...}

    """
    if isinstance(model, list) and len(model) == 1:
        model = model[0]

    res = model.modelfit_results
    if not res:
        raise ValueError('No modelfit results available')

    model.parameters = res.parameter_estimates

    if model.initial_individual_estimates is not None or force_individual_estimates:
        model.initial_individual_estimates = res.individual_estimates

    return model



# The modeling features language
# A high level language to describe model features and ranges of model features


# ; and \n separates features
# feature names are case insensitive
# FEATURE_NAME(options)     options is a comma separated list
# absorption(x)   - x can be FO, ZO, SEQ-ZO-FO, [FO, ZO] for multiple or * for all
# elimination(x)  - x can be FO, ZO, MM and MIX-FO-MM
# peripherals(n)  - n can be 0, 1, ... or a..b for a range (inclusive in both ends)
# transits(n)     - n as above


import functools
import itertools

from lark import Lark
from lark.visitors import Interpreter

import pharmpy.modeling as modeling

grammar = r"""
start: feature (_SEPARATOR feature)*
feature: absorption | elimination | peripherals | transits | lagtime

absorption: "ABSORPTION"i "(" (option) ")"
elimination: "ELIMINATION"i "(" (option) ")"
peripherals: "PERIPHERALS"i "(" (option) ")"
transits: "TRANSITS"i "(" (option) ")"
lagtime: "LAGTIME()"
option: (wildcard | range | value | array)
wildcard: "*"
range: NUMBER ".." NUMBER
value: /[a-zA-Z0-9-]+/
array: "[" [value ("," value)*] "]"
_SEPARATOR: /;|\n/
NUMBER: /\d+/
%ignore " "
"""


class ModelFeature:
    pass


class Absorption(ModelFeature):
    def __init__(self, tree):
        self.args = OneArgInterpreter('absorption', ['FO', 'ZO', 'SEQ-ZO-FO']).interpret(tree)
        self._funcs = dict()
        for arg in self.args:
            name = f'ABSORPTION({arg})'
            if arg == 'FO':
                self._funcs[name] = modeling.set_first_order_absorption
            elif arg == 'ZO':
                self._funcs[name] = modeling.set_zero_order_absorption
            elif arg == 'SEQ-ZO-FO':
                self._funcs[name] = modeling.set_seq_zo_fo_absorption
            else:
                raise ValueError(f'Absorption {arg} not supported')


class Elimination(ModelFeature):
    def __init__(self, tree):
        self.args = OneArgInterpreter('elimination', ['FO', 'ZO', 'MM', 'MIX-FO-MM']).interpret(
            tree
        )
        self._funcs = dict()
        for arg in self.args:
            name = f'ELIMINATION({arg})'
            if arg == 'FO':
                self._funcs[name] = modeling.set_first_order_elimination
            elif arg == 'ZO':
                self._funcs[name] = modeling.set_zero_order_elimination
            elif arg == 'MM':
                self._funcs[name] = modeling.set_michaelis_menten_elimination
            elif arg == 'MIX-FO-MM':
                self._funcs[name] = modeling.set_mixed_mm_fo_elimination
            else:
                raise ValueError(f'Elimination {arg} not supported')


class Transits(ModelFeature):
    def __init__(self, tree):
        self.args = OneArgInterpreter('transits', []).interpret(tree)
        self._funcs = dict()
        for arg in self.args:
            name = f'TRANSITS({arg})'
            self._funcs[name] = functools.partial(modeling.set_transit_compartments, n=arg)


class Peripherals(ModelFeature):
    def __init__(self, tree):
        self.args = OneArgInterpreter('peripherals', []).interpret(tree)
        self._funcs = dict()
        for arg in self.args:
            name = f'PERIPHERALS({arg})'
            self._funcs[name] = functools.partial(modeling.set_peripheral_compartments, n=arg)


class Lagtime(ModelFeature):
    def __init__(self, tree):
        self._funcs = {'LAGTIME()': modeling.add_lag_time}
        self.args = None


class OneArgInterpreter(Interpreter):
    def __init__(self, name, a):
        self.name = name
        self.all = a

    def visit_children(self, tree):
        a = super().visit_children(tree)
        return set().union(*a)

    def feature(self, tree):
        return self.visit_children(tree)

    def absorption(self, tree):
        if self.name == 'absorption':
            return self.visit_children(tree)
        else:
            return []

    def elimination(self, tree):
        if self.name == 'elimination':
            return self.visit_children(tree)
        else:
            return []

    def transits(self, tree):
        if self.name == 'transits':
            return self.visit_children(tree)
        else:
            return []

    def peripherals(self, tree):
        if self.name == 'peripherals':
            return self.visit_children(tree)
        else:
            return []

    def option(self, tree):
        return self.visit_children(tree)

    def value(self, tree):
        value = tree.children[0].value
        if self.name == 'transits' or self.name == 'peripherals':
            return {int(value)}
        else:
            return {value.upper()}

    def array(self, tree):
        return self.visit_children(tree)

    def wildcard(self, tree):
        if self.name == 'peripherals' or self.name == 'transits':
            raise ValueError(f'Wildcard (*) not supported for {self.name}')
        return set(self.all)

    def range(self, tree):
        return set(range(int(tree.children[0]), int(tree.children[1]) + 1))


class ModelFeatures:
    def __init__(self, code):
        parser = Lark(grammar)
        tree = parser.parse(code)
        self._all_features = []
        if list(tree.find_data('absorption')):
            self.absorption = Absorption(tree)
            self._all_features.append(self.absorption)
        if list(tree.find_data('elimination')):
            self.elimination = Elimination(tree)
            self._all_features.append(self.elimination)
        if list(tree.find_data('transits')):
            self.transits = Transits(tree)
            self._all_features.append(self.transits)
        if list(tree.find_data('peripherals')):
            self.peripherals = Peripherals(tree)
            self._all_features.append(self.peripherals)
        if list(tree.find_data('lagtime')):
            self.lagtime = Lagtime(tree)
            self._all_features.append(self.lagtime)

    def all_funcs(self):
        funcs = dict()
        for feat in self._all_features:
            funcs.update(feat._funcs)
        return funcs

    def next_funcs(self, have):
        funcs = dict()
        names = [s.split('(')[0].strip().upper() for s in have]
        for feat, func in self.all_funcs().items():
            curname = feat.split('(')[0].strip().upper()
            if curname not in names:
                funcs[feat] = func
        return funcs

    def all_combinations(self):
        feats = []
        for feat in self._all_features:
            feats.append([None] + list(feat._funcs.keys()))
        for t in itertools.product(*feats):
            a = [elt for elt in t if elt is not None]
            if a:
                yield a

    def get_funcs_same_type(self, feat):
        if hasattr(self, 'absorption') and feat in self.absorption._funcs.keys():
            funcs = self.absorption._funcs.values()
        elif hasattr(self, 'elimination') and feat in self.elimination._funcs.keys():
            funcs = self.elimination._funcs.values()
        elif hasattr(self, 'transits') and feat in self.transits._funcs.keys():
            funcs = self.transits._funcs.values()
        elif hasattr(self, 'peripherals') and feat in self.peripherals._funcs.keys():
            funcs = self.peripherals._funcs.values()
        elif hasattr(self, 'lagtime') and feat in self.lagtime._funcs.keys():
            funcs = self.lagtime._funcs.values()
        else:
            return []
        return list(funcs)



from io import StringIO

import pytest
import sympy

import pharmpy.symbols
from pharmpy import Model
from pharmpy.plugins.nonmem.advan import compartmental_model


def S(x):
    return pharmpy.symbols.symbol(x)


@pytest.mark.parametrize(
    'advan,trans,compmat,amounts,strodes,corrics',
    [
        (
            'ADVAN1',
            'TRANS1',
            [[-S('K'), 0], [S('K'), 0]],
            [S('A_CENTRAL'), S('A_OUTPUT')],
            [
                'Eq(Derivative(A_CENTRAL(t), t), -K*A_CENTRAL(t))',
                'Eq(Derivative(A_OUTPUT(t), t), K*A_CENTRAL(t))',
            ],
            {sympy.Function('A_CENTRAL')(0): S('AMT'), sympy.Function('A_OUTPUT')(0): 0},
        ),
        (
            'ADVAN1',
            'TRANS2',
            [[-S('CL') / S('V'), 0], [S('CL') / S('V'), 0]],
            [S('A_CENTRAL'), S('A_OUTPUT')],
            [
                'Eq(Derivative(A_CENTRAL(t), t), -CL*A_CENTRAL(t)/V)',
                'Eq(Derivative(A_OUTPUT(t), t), CL*A_CENTRAL(t)/V)',
            ],
            {sympy.Function('A_CENTRAL')(0): S('AMT'), sympy.Function('A_OUTPUT')(0): 0},
        ),
        (
            'ADVAN2',
            'TRANS1',
            [[-S('KA'), 0, 0], [S('KA'), -S('K'), 0], [0, S('K'), 0]],
            [S('A_DEPOT'), S('A_CENTRAL'), S('A_OUTPUT')],
            [
                'Eq(Derivative(A_DEPOT(t), t), -KA*A_DEPOT(t))',
                'Eq(Derivative(A_CENTRAL(t), t), -K*A_CENTRAL(t) + KA*A_DEPOT(t))',
                'Eq(Derivative(A_OUTPUT(t), t), K*A_CENTRAL(t))',
            ],
            {
                sympy.Function('A_DEPOT')(0): S('AMT'),
                sympy.Function('A_CENTRAL')(0): 0,
                sympy.Function('A_OUTPUT')(0): 0,
            },
        ),
        (
            'ADVAN2',
            'TRANS2',
            [[-S('KA'), 0, 0], [S('KA'), -S('CL') / S('V'), 0], [0, S('CL') / S('V'), 0]],
            [S('A_DEPOT'), S('A_CENTRAL'), S('A_OUTPUT')],
            [
                'Eq(Derivative(A_DEPOT(t), t), -KA*A_DEPOT(t))',
                'Eq(Derivative(A_CENTRAL(t), t), -CL*A_CENTRAL(t)/V + KA*A_DEPOT(t))',
                'Eq(Derivative(A_OUTPUT(t), t), CL*A_CENTRAL(t)/V)',
            ],
            {
                sympy.Function('A_DEPOT')(0): S('AMT'),
                sympy.Function('A_CENTRAL')(0): 0,
                sympy.Function('A_OUTPUT')(0): 0,
            },
        ),
        (
            'ADVAN3',
            'TRANS1',
            [[-S('K12') - S('K'), S('K21'), 0], [S('K12'), -S('K21'), 0], [S('K'), 0, 0]],
            [S('A_CENTRAL'), S('A_PERIPHERAL'), S('A_OUTPUT')],
            [
                'Eq(Derivative(A_CENTRAL(t), t), K21*A_PERIPHERAL(t) + (-K - K12)*A_CENTRAL(t))',
                'Eq(Derivative(A_PERIPHERAL(t), t), K12*A_CENTRAL(t) - K21*A_PERIPHERAL(t))',
                'Eq(Derivative(A_OUTPUT(t), t), K*A_CENTRAL(t))',
            ],
            {
                sympy.Function('A_CENTRAL')(0): S('AMT'),
                sympy.Function('A_PERIPHERAL')(0): 0,
                sympy.Function('A_OUTPUT')(0): 0,
            },
        ),
        (
            'ADVAN4',
            'TRANS1',
            [
                [-S('KA'), 0, 0, 0],
                [S('KA'), -S('K23') - S('K'), S('K32'), 0],
                [0, S('K23'), -S('K32'), 0],
                [0, S('K'), 0, 0],
            ],
            [S('A_DEPOT'), S('A_CENTRAL'), S('A_PERIPHERAL'), S('A_OUTPUT')],
            [
                'Eq(Derivative(A_DEPOT(t), t), -KA*A_DEPOT(t))',
                'Eq(Derivative(A_CENTRAL(t), t), K32*A_PERIPHERAL(t) + '
                'KA*A_DEPOT(t) + (-K - K23)*A_CENTRAL(t))',
                'Eq(Derivative(A_PERIPHERAL(t), t), K23*A_CENTRAL(t) - K32*A_PERIPHERAL(t))',
                'Eq(Derivative(A_OUTPUT(t), t), K*A_CENTRAL(t))',
            ],
            {
                sympy.Function('A_DEPOT')(0): S('AMT'),
                sympy.Function('A_CENTRAL')(0): 0,
                sympy.Function('A_PERIPHERAL')(0): 0,
                sympy.Function('A_OUTPUT')(0): 0,
            },
        ),
        (
            'ADVAN10',
            'TRANS1',
            [
                [-S('VM') / (S('KM') + sympy.Function('A_CENTRAL')(S('t'))), 0],
                [S('VM') / (S('KM') + sympy.Function('A_CENTRAL')(S('t'))), 0],
            ],
            [S('A_CENTRAL'), S('A_OUTPUT')],
            [
                'Eq(Derivative(A_CENTRAL(t), t), -VM*A_CENTRAL(t)/(KM + A_CENTRAL(t)))',
                'Eq(Derivative(A_OUTPUT(t), t), VM*A_CENTRAL(t)/(KM + A_CENTRAL(t)))',
            ],
            {sympy.Function('A_CENTRAL')(0): S('AMT'), sympy.Function('A_OUTPUT')(0): 0},
        ),
        (
            'ADVAN11',
            'TRANS1',
            [
                [-S('K12') - S('K13') - S('K'), S('K21'), S('K31'), 0],
                [S('K12'), -S('K21'), 0, 0],
                [S('K13'), 0, -S('K31'), 0],
                [S('K'), 0, 0, 0],
            ],
            [S('A_CENTRAL'), S('A_PERIPHERAL1'), S('A_PERIPHERAL2'), S('A_OUTPUT')],
            [
                'Eq(Derivative(A_CENTRAL(t), t), K21*A_PERIPHERAL1(t) + K31*A_PERIPHERAL2(t) + '
                '(-K - K12 - K13)*A_CENTRAL(t))',
                'Eq(Derivative(A_PERIPHERAL1(t), t), K12*A_CENTRAL(t) - K21*A_PERIPHERAL1(t))',
                'Eq(Derivative(A_PERIPHERAL2(t), t), K13*A_CENTRAL(t) - K31*A_PERIPHERAL2(t))',
                'Eq(Derivative(A_OUTPUT(t), t), K*A_CENTRAL(t))',
            ],
            {
                sympy.Function('A_CENTRAL')(0): S('AMT'),
                sympy.Function('A_PERIPHERAL1')(0): 0,
                sympy.Function('A_PERIPHERAL2')(0): 0,
                sympy.Function('A_OUTPUT')(0): 0,
            },
        ),
        (
            'ADVAN12',
            'TRANS1',
            [
                [-S('KA'), 0, 0, 0, 0],
                [S('KA'), -S('K23') - S('K24') - S('K'), S('K32'), S('K42'), 0],
                [0, S('K23'), -S('K32'), 0, 0],
                [0, S('K24'), 0, -S('K42'), 0],
                [0, S('K'), 0, 0, 0],
            ],
            [S('A_DEPOT'), S('A_CENTRAL'), S('A_PERIPHERAL1'), S('A_PERIPHERAL2'), S('A_OUTPUT')],
            [
                'Eq(Derivative(A_DEPOT(t), t), -KA*A_DEPOT(t))',
                'Eq(Derivative(A_CENTRAL(t), t), K32*A_PERIPHERAL1(t) + '
                'K42*A_PERIPHERAL2(t) + KA*A_DEPOT(t) + (-K - K23 - K24)*A_CENTRAL(t))',
                'Eq(Derivative(A_PERIPHERAL1(t), t), K23*A_CENTRAL(t) - K32*A_PERIPHERAL1(t))',
                'Eq(Derivative(A_PERIPHERAL2(t), t), K24*A_CENTRAL(t) - K42*A_PERIPHERAL2(t))',
                'Eq(Derivative(A_OUTPUT(t), t), K*A_CENTRAL(t))',
            ],
            {
                sympy.Function('A_DEPOT')(0): S('AMT'),
                sympy.Function('A_CENTRAL')(0): 0,
                sympy.Function('A_PERIPHERAL1')(0): 0,
                sympy.Function('A_PERIPHERAL2')(0): 0,
                sympy.Function('A_OUTPUT')(0): 0,
            },
        ),
    ],
)
def test_pheno(pheno_path, advan, trans, compmat, amounts, strodes, corrics):
    model = Model(pheno_path)
    cm, ass = compartmental_model(model, advan, trans)

    assert ass.symbol == S('F')
    assert ass.expression == S('A_CENTRAL') / S('S1')
    assert cm.compartmental_matrix == sympy.Matrix(compmat)
    assert cm.amounts == sympy.Matrix(amounts)
    exodes = cm.to_explicit_system()
    odes, ics = exodes.odes, exodes.ics
    odes = [str(ode) for ode in odes]
    assert odes == strodes
    assert ics == corrics


def test_advan5(testdata):
    model = Model(testdata / 'nonmem' / 'DDMODEL00000130')
    cm, ass = compartmental_model(model, 'ADVAN5', 'TRANS1')
    assert ass.symbol == S('F')
    assert ass.expression == S('A_CMS1')
    assert cm.amounts == sympy.Matrix(
        [S('A_CMS1'), S('A_CMS2'), S('A_COL1'), S('A_INTM'), S('A_INTM2'), S('A_OUTPUT')]
    )
    compmat = sympy.Matrix(
        [
            [-S('K12') - S('K10') - S('K14'), S('K21'), S('K31'), 0, 0, 0],
            [S('K12'), -S('K21') - S('K25'), 0, 0, 0, 0],
            [0, 0, -S('K31') - S('K30'), S('K43'), 0, 0],
            [S('K14'), 0, 0, -S('K43') - S('K40') - S('K45'), S('K54'), 0],
            [0, S('K25'), 0, S('K45'), -S('K54'), 0],
            [S('K10'), 0, S('K30'), S('K40'), 0, 0],
        ]
    )
    assert cm.compartmental_matrix == compmat


def test_rate_constants():
    code = """$PROBLEM
$INPUT ID VISI DAT1 DGRP DOSE FLAG ONO XIME DVO NEUY SCR AGE SEX
       NYHA WT COMP IACE DIG DIU NUMB TAD TIME VID CRCL AMT SS II
       VID1 CMT CONO DV EVID OVID SHR SHR2
$DATA mx19B.csv IGNORE=@
$SUBROUTINE ADVAN5 TRANS1
$MODEL COMPARTMENT=(TRANSIT1 DEFDOSE) COMPARTMENT=(TRANSIT2) COMPARTMENT=(TRANSIT3)
COMPARTMENT=(TRANSIT4) COMPARTMENT=(TRANSIT5) COMPARTMENT=(TRANSIT6)
COMPARTMENT=(TRANSIT7) COMPARTMENT=(TRANSIT8) COMPARTMENT=(DEPOT) COMPARTMENT=(CENTRAL)
$PK
MDT = THETA(4)*EXP(ETA(4))
CL = THETA(1) * EXP(ETA(1))
VC = THETA(2) * EXP(ETA(2))
MAT = THETA(3) * EXP(ETA(3))
KA = 1/MAT
V = VC
K12 = 8/MDT
K910 = KA
K100 = CL/V
K23 = 8/MDT
K34 = 8/MDT
K45 = 8/MDT
K56 = 8/MDT
K67 = 8/MDT
K78 = 8/MDT
K89 = 8/MDT
$ERROR
CONC = A(10)/VC
Y = CONC + CONC * EPS(1)
$ESTIMATION METHOD=1 INTER MAXEVAL=9999
$COVARIANCE PRINT=E
$THETA  (0,22.7,Inf) ; POP_CL
$THETA  (0,128,Inf) ; POP_VC
$THETA  (0,0.196,Inf) ; POP_MAT
$THETA  (0,0.003) ; POP_MDT
$OMEGA  0.001  ;     IIV_CL
$OMEGA  0.001  ;     IIV_VC
$OMEGA  0.001  ;    IIV_MAT
$OMEGA  0.001 ; IIV_MDT
$SIGMA  0.273617  ;   RUV_PROP
"""
    model = Model(StringIO(code))
    odes = model.statements.ode_system
    assert odes.get_flow(odes.central_compartment, odes.output_compartment) == sympy.Symbol('K100')



import warnings
from pathlib import Path

import numpy as np
import pandas as pd
import scipy.stats

import pharmpy.visualization
from pharmpy.model_factory import Model
from pharmpy.results import ModelfitResults, Results
from pharmpy.tools.psn_helpers import cmd_line_model_path, model_paths


class BootstrapResults(Results):
    rst_path = Path(__file__).parent / 'report.rst'

    # FIXME: Should inherit from results that take multiple runs like bootstrap, cdd etc.
    def __init__(
        self,
        parameter_statistics=None,
        parameter_distribution=None,
        covariance_matrix=None,
        ofv_distribution=None,
        ofv_statistics=None,
        included_individuals=None,
        ofvs=None,
        parameter_estimates=None,
        ofv_plot=None,
        parameter_estimates_correlation_plot=None,
        dofv_quantiles_plot=None,
        parameter_estimates_histogram=None,
    ):
        self.parameter_statistics = parameter_statistics
        self.parameter_distribution = parameter_distribution
        self.covariance_matrix = covariance_matrix
        self.ofv_distribution = ofv_distribution
        self.ofv_statistics = ofv_statistics
        self.included_individuals = included_individuals
        self.ofvs = ofvs
        self.parameter_estimates = parameter_estimates
        self.ofv_plot = ofv_plot
        self.parameter_estimates_correlation_plot = parameter_estimates_correlation_plot
        self.dofv_quantiles_plot = dofv_quantiles_plot
        self.parameter_estimates_histogram = parameter_estimates_histogram

    def add_plots(self):
        self.ofv_plot = self.plot_ofv()
        self.parameter_estimates_correlation_plot = self.plot_parameter_estimates_correlation()
        self.dofv_quantiles_plot = self.plot_dofv_quantiles()
        self.parameter_estimates_histogram = self.plot_parameter_estimates_histogram()

    def plot_ofv(self):
        plot = pharmpy.visualization.histogram(
            self.ofvs['bootstrap_bootdata_ofv'], title='Bootstrap OFV'
        )
        return plot

    def plot_dofv_quantiles(self):
        ofvs = self.ofvs
        dofvs = ofvs['delta_bootdata'].sort_values().reset_index(drop=True)
        dofvs_boot_base = ofvs['delta_origdata'].sort_values().reset_index(drop=True)
        quantiles = np.linspace(0.0, 1.0, num=len(dofvs))
        degrees = len(self.parameter_distribution)
        chi2_dist = scipy.stats.chi2(df=degrees)
        chi2 = chi2_dist.ppf(quantiles)
        degrees_dofvs = self.ofv_statistics['mean']['delta_bootdata']
        degrees_boot_base = self.ofv_statistics['mean']['delta_origdata']
        df_dict = {'quantiles': quantiles, f'Reference χ²({degrees})': chi2}
        if not np.isnan(degrees_dofvs):
            df_dict[
                (
                    'Original model OFV - Bootstrap model OFV (both using bootstrap datasets)',
                    f'Estimated df = {degrees_dofvs:.2f}',
                )
            ] = dofvs
        if not np.isnan(degrees_boot_base):
            df_dict[
                (
                    'Bootstrap model OFV - Original model OFV (both using original dataset)',
                    f'Estimated df = {degrees_boot_base:.2f}',
                )
            ] = dofvs_boot_base
        df = pd.DataFrame(df_dict)
        plot = pharmpy.visualization.line_plot(
            df,
            'quantiles',
            xlabel='Distribution quantiles',
            ylabel='dOFV',
            legend_title='Distribution',
            title='dOFV quantiles',
        )
        return plot

    def plot_parameter_estimates_correlation(self):
        pe = self.parameter_estimates
        plot = pharmpy.visualization.scatter_matrix(pe)
        return plot

    def plot_parameter_estimates_histogram(self):
        pe = self.parameter_estimates
        plot = pharmpy.visualization.facetted_histogram(pe)
        return plot


def calculate_results(
    bootstrap_models, original_model=None, included_individuals=None, dofv_results=None
):
    results = [m.modelfit_results for m in bootstrap_models if m.modelfit_results is not None]
    if original_model:
        original_results = original_model.modelfit_results
    else:
        original_results = None

    if original_results is None:
        warnings.warn(
            'No results for the base model could be read. Cannot calculate bias and '
            'original_bootdata_ofv'
        )

    df = pd.DataFrame()
    for res in results:
        df = df.append(res.parameter_estimates, ignore_index=True, sort=False)
    df = df.reindex(results[0].parameter_estimates.index, axis=1)
    parameter_estimates = df.reset_index(drop=True)

    covariance_matrix = df.cov()

    df = parameter_estimates.copy()
    mean = df.mean()
    if original_results is not None:
        orig = original_results.parameter_estimates
        bias = mean - orig
    else:
        bias = np.nan
    statistics = pd.DataFrame(
        {'mean': mean, 'median': df.median(), 'bias': bias, 'stderr': df.std()}
    )
    statistics['RSE'] = statistics['stderr'] / statistics['mean']

    df = parameter_estimates
    distribution = create_distribution(df)

    boot_ofvs = [x.ofv for x in results]
    ofvs = pd.Series(boot_ofvs, name='bootstrap_bootdata_ofv')
    ofvs = pd.DataFrame(ofvs)
    ofvs['original_bootdata_ofv'] = np.nan
    ofvs['bootstrap_origdata_ofv'] = np.nan
    if original_results is not None:
        base_iofv = original_results.individual_ofv
        if included_individuals and base_iofv is not None:
            for i, included in enumerate(included_individuals):
                base_ofv = base_iofv[included].sum()
                ofvs.at[i, 'original_bootdata_ofv'] = base_ofv

    if dofv_results is not None:
        for i, res in enumerate(dofv_results):
            if res is not None:
                ofvs.at[i, 'bootstrap_origdata_ofv'] = res.ofv

    if original_results is not None:
        base_ofv = original_results.ofv
        ofvs['original_origdata_ofv'] = base_ofv

    ofvs['delta_bootdata'] = ofvs['original_bootdata_ofv'] - ofvs['bootstrap_bootdata_ofv']

    if original_results is not None:
        ofvs['delta_origdata'] = ofvs['bootstrap_origdata_ofv'] - base_ofv
    else:
        ofvs['delta_origdata'] = np.nan

    with warnings.catch_warnings():
        # Catch numpy warnings beause of NaN in ofvs
        warnings.filterwarnings('ignore', r'All-NaN slice encountered')
        warnings.filterwarnings('ignore', 'Mean of empty slice')
        ofv_dist = create_distribution(ofvs)
        ofv_stats = pd.DataFrame(
            {'mean': ofvs.mean(), 'median': ofvs.median(), 'stderr': ofvs.std()}
        )

    res = BootstrapResults(
        covariance_matrix=covariance_matrix,
        parameter_statistics=statistics,
        parameter_distribution=distribution,
        ofv_distribution=ofv_dist,
        ofv_statistics=ofv_stats,
        included_individuals=included_individuals,
        ofvs=ofvs,
        parameter_estimates=parameter_estimates,
    )

    return res


def create_distribution(df):
    dist = pd.DataFrame(
        {
            'min': df.min(),
            '0.05%': df.quantile(0.0005),
            '0.5%': df.quantile(0.005),
            '2.5%': df.quantile(0.025),
            '5%': df.quantile(0.05),
            'median': df.median(),
            '95%': df.quantile(0.95),
            '97.5%': df.quantile(0.975),
            '99.5%': df.quantile(0.995),
            '99.95%': df.quantile(0.9995),
            'max': df.max(),
        }
    )
    return dist


def psn_bootstrap_results(path):
    """Create bootstrapresults from a PsN bootstrap run

    :param path: Path to PsN boostrap run directory
    :return: A :class:`BootstrapResults` object
    """
    path = Path(path)

    models = [Model(p) for p in model_paths(path, 'bs_pr1_*.mod')]
    # Read the results already now to give an appropriate error if no results exists
    results = [m.modelfit_results for m in models if m.modelfit_results is not None]
    if not results:
        raise FileNotFoundError("No model results available in m1")
    try:
        base_model = Model(cmd_line_model_path(path))
    except FileNotFoundError:
        base_model = None

    # Read dOFV results in NONMEM specific way. Models have multiple $PROBLEM
    # Create proper result objects to pass to calculate_results
    dofv_results = None
    if (path / 'm1' / 'dofv_1.mod').is_file():
        from pharmpy.plugins.nonmem.table import NONMEMTableFile

        dofv_results = []
        for table_path in (path / 'm1').glob('dofv_*.ext'):
            table_file = NONMEMTableFile(table_path)
            next_table = 1
            for table in table_file:
                while next_table != table.number:
                    dofv_results.append(None)
                    next_table += 1
                res = ModelfitResults(ofv=table.final_ofv)
                dofv_results.append(res)
                next_table += 1

    incinds = pd.read_csv(path / 'included_individuals1.csv', header=None).values.tolist()
    res = calculate_results(
        models, original_model=base_model, included_individuals=incinds, dofv_results=dofv_results
    )
    return res



"""Factory for creating a pharmpy model from a "native" object representation.

Definitions
===========
"""

import io
import pathlib
from pathlib import Path

from pharmpy.plugins.utils import detect_model


def Model(obj, **kwargs):
    """Factory for creating a :class:`pharmpy.model` object from an object representing the model
    (i.e. path).

    .. _path-like object: https://docs.python.org/3/glossary.html#term-path-like-object

    Parameters
    ----------
    obj
        Currently a `path-like object`_ pointing to the model file.

    Returns
    -------
    - Generic :class:`~pharmpy.generic.Model` if path is None, otherwise appropriate implementation
      is invoked (e.g. NONMEM7 :class:`~pharmpy.api_nonmem.model.Model`).
    """
    if isinstance(obj, str):
        path = Path(obj)
    elif isinstance(obj, pathlib.Path):
        path = obj
    elif isinstance(obj, io.IOBase):
        path = None
    else:
        raise ValueError("Unknown input type to Model constructor")
    if path is not None:
        with open(path, 'r', encoding='latin-1') as fp:
            code = fp.read()
    else:
        code = obj.read()
    model_class = detect_model(code)
    model = model_class(code, path, **kwargs)
    # Setup model database here
    # Read in model results here?
    # Set filename extension?
    return model



# Read dataset from file
import re
import warnings
from io import StringIO
from pathlib import Path

import numpy as np
import pandas as pd
from lark import Lark

import pharmpy.data
from pharmpy.data import ColumnType, DatasetError, DatasetWarning


class NMTRANDataIO(StringIO):
    """An IO class that is a prefilter for pandas.read_table.
    Things that must be done before using pandas will be done here.
    Currently it takes care of filtering out ignored rows and handles special delimiter cases
    """

    def __init__(self, filename_or_io, ignore_character='#'):
        """filename_or_io is a string with a path, a path object or any IO object, i.e. StringIO"""
        if not ignore_character:
            ignore_character = '#'
        if hasattr(filename_or_io, 'read'):
            contents = filename_or_io.read()
        else:
            with open(str(filename_or_io), 'r', encoding='latin-1') as datafile:
                contents = datafile.read()  # All variations of newlines are converted into \n

        if ignore_character == '@':
            # FIXME: Does this really handle the final line with no new line?
            comment_regexp = re.compile(r'^[ \t]*[A-Za-z#@].*\n', re.MULTILINE)
        else:
            comment_regexp = re.compile('^[' + ignore_character + '].*\n', re.MULTILINE)
        contents = re.sub(comment_regexp, '', contents)

        if re.search(r' \t', contents):  # Space before TAB not allowed (see documentation)
            raise DatasetError(
                "The dataset contains a TAB preceeded by a space, "
                "which is not allowed by NM-TRAN"
            )

        if re.search(r'^[ \t]*\n$', contents, re.MULTILINE):  # Blank lines
            raise DatasetError(
                "The dataset contains one or more blank lines. This is not "
                "allowed by NM-TRAN without the BLANKOK option"
            )

        super().__init__(contents)


def convert_fortran_number(number_string):
    """This function will try to convert the number_string from the general fortran exponential format
    into an np.float64. It covers "1d1", "1D1", "a+b", "a-b", "+" and "-". All other cases will
    return None to signal that the number_string is not of the special form.

    Move somewhere else. Will be used in output parsing as well
    """
    try:
        y = np.float64(number_string)
        return y
    except (TypeError, ValueError):
        pass

    if number_string == '+' or number_string == '-':
        return 0.0

    m = re.match(r'([+\-]?)([^+\-dD]*)([+-])([^+\-dD]*)', number_string)
    if m:
        mantissa_sign = '-' if m.group(1) == '-' else ''
        mantissa = m.group(2)
        exponent_sign = m.group(3)
        exponent = m.group(4)
        return np.float64(mantissa_sign + mantissa + "E" + exponent_sign + exponent)

    if "D" in number_string or "d" in number_string:
        clean_number = number_string.replace("D", "e").replace("d", "e")
        try:
            y = np.float64(clean_number)
            return y
        except (TypeError, ValueError):
            pass

    raise ValueError(f"Could not convert the fortran number {number_string} to float")


def _convert_data_item(x, null_value):
    if x is None or x == '.' or x == '':
        x = null_value
    if len(x) > 24:
        raise DatasetError("The dataset contains an item that is longer than 24 characters")
    try:
        converted = convert_fortran_number(x)
    except ValueError as e:
        raise DatasetError(str(e)) from e
    if converted in pharmpy.data.conf.na_values:
        return np.nan
    return converted


def _make_ids_unique(df, columns):
    """Check if id numbers are reused and make renumber. If not simply pass through the dataset."""
    try:
        id_label = df.pharmpy.id_label
    except KeyError:
        return df
    if id_label in columns:
        id_series = df[id_label]
        id_change = id_series.diff(1) != 0
        if len(id_series[id_change]) != len(df.pharmpy.ids):
            warnings.warn(
                "Dataset contains non-unique id numbers. Renumbering starting from 1",
                DatasetWarning,
            )
            df[df.pharmpy.id_label] = id_change.cumsum()
    return df


def _filter_ignore_accept(df, ignore, accept, null_value):
    if ignore and accept:
        raise ValueError("Cannot have both IGNORE and ACCEPT")
    if not ignore and not accept:
        return df

    if ignore:
        statements = ignore
    else:
        statements = accept
    grammar = r'''
        start: column [space] operator [space] value | column [space] value
        column: COLNAME
        operator: OP_EQ | OP_STR_EQ | OP_NE | OP_STR_NE | OP_LT | OP_GT | OP_LT_EQ | OP_GT_EQ
        value: TEXT | QUOTE
        space: WS
        COLNAME: /\w+/
        WS: /\s+/
        OP_EQ    : ".EQN."
        OP_STR_EQ: ".EQ." | "==" | "="
        OP_NE    : ".NEN."
        OP_STR_NE: ".NE." | "/="
        OP_LT    : ".LT." | "<"
        OP_GT    : ".GT." | ">"
        OP_LT_EQ : ".LE." | "<="
        OP_GT_EQ : ".GE." | ">="
        TEXT: /[^"',;()=\s]+/
        QUOTE: /"[^"]*"/
             | /'[^']*'/
    '''
    parser = Lark(grammar)
    for s in statements:
        tree = parser.parse(s)
        operator = '=='
        operator_type = str
        for st in tree.iter_subtrees():
            if st.data == 'column':
                column = str(st.children[0])
            elif st.data == 'value':
                value = str(st.children[0])
            elif st.data == 'operator':
                operator_token = st.children[0]
                tp = operator_token.type
                if tp == 'OP_EQ':
                    operator = '=='
                    operator_type = float
                elif tp == 'OP_NE':
                    operator = '!='
                    operator_type = float
                elif tp == 'OP_LT':
                    operator = '<'
                    operator_type = float
                elif tp == 'OP_GT':
                    operator = '>'
                    operator_type = float
                elif tp == 'OP_LT_EQ':
                    operator = '<='
                    operator_type = float
                elif tp == 'OP_GT_EQ':
                    operator = '>='
                    operator_type = float
                elif tp == 'OP_STR_EQ':
                    operator = '=='
                    operator_type = str
                elif tp == 'OP_STR_NE':
                    operator = '!='
                    operator_type = str
        if len(value) >= 3 and (
            (value.startswith("'") and value.endswith("'"))
            or (value.startswith('"') and value.endswith('"'))
        ):
            value = value[1:-1]

        if operator_type == str:
            expression = f'{column} {operator} "{value}"'
            if ignore:
                expression = 'not(' + expression + ')'
            df.query(expression, inplace=True)
        else:
            # Need to temporary convert column. Refer to NONMEM fileformat documentation
            # for further information.
            # Using a name with spaces since this cannot collide with other NONMEM names
            magic_colname = 'a a'
            df[magic_colname] = df[column].apply(_convert_data_item, args=(str(null_value),))
            expression = f'`{magic_colname}` {operator} {value}'
            if ignore:
                expression = 'not(' + expression + ')'
            df.query(expression, inplace=True)
            df.drop(labels=magic_colname, axis=1, inplace=True)
    df.reset_index(drop=True, inplace=True)
    return df


def infer_column_type(colname):
    """If possible infer the column type from the column name else use unknown"""
    if colname == 'ID' or colname == 'L1':
        return ColumnType.ID
    elif colname == 'DV':
        return ColumnType.DV
    elif colname == 'MDV':
        return ColumnType.EVENT
    else:
        return ColumnType.UNKNOWN


# def _translate_nonmem_time_value(time):
#    if ':' in time:
#        components = time.split(':')
#        if len(components) > 3:
#            raise DatasetError(f'Bad TIME format: {time}')
#        hours = convert_fortran_number(components[0]) + convert_fortran_number(components[1]) * 60
#        if len(components) == 3:
#            hours += convert_fortran_number(components[2])
#        return hours
#    else:
#        return time


# def _translate_nonmem_time_and_date(df, timecol='TIME', datecol=None):
#    relative_time = df[timecol].str.contains(':').any()
#    if relative_time:
#        df[timecol] = df[timecol].apply(_translate_nonmem_time_value)
#        df[timecol] = df[timecol] - df[timecol].first()
#    return df
#    """
#    if date:
#        m = re.match(r'(?P<first>^[-]?\d+)(\D+(?P<second>\d+))?(\D+(?P<third>\d+))?$', date)
#        if not m:
#            raise ValueError('Bad DATE format: {date}')

#        first = m.group('first')
#        second = m.group('second')
#        third = m.group('third')
#        if second is None and third is None:
#            hours += float(first) * 24      # day
#        elif third is None:
#            float(first) * 24 + float(second)       # day and month
#        elif  date_label == 'DAT2':        # yy-mm-dd
#    """


def read_nonmem_dataset(
    path_or_io,
    raw=False,
    ignore_character='#',
    colnames=tuple(),
    coltypes=None,
    drop=None,
    null_value='0',
    parse_columns=tuple(),
    ignore=None,
    accept=None,
):
    """Read a nonmem dataset from file
     column types will be inferred from the column names

    raw - minimal processing, data will be kept in string format.
    ignore_character
    colnames - List or tuple of names to give each column given in order. Names need to be unique
    drop - A list or tuple of booleans of which columns to drop
    null_value - Value to use for NULL, i.e. empty records or padding
    parse_columns - Only applicable when raw=True. A list of columns to parse.
    ignore/accept - List of ignore/accept expressions

     The following postprocessing operations are done to a non-raw dataset
     1. Convert ordinary floating point numbers to float64
     2. Convert numbers of special fortran format to float64
     3. Convert None, '.', empty string to the NULL value
     4. Convert Inf/NaN properly
     5. Pad with null_token columns if $INPUT has more columns than the dataset
     6. Strip away superfluous columns from the dataset
    """
    if drop is None:
        drop = [False] * len(colnames)

    non_dropped = [name for name, dropped in zip(colnames, drop) if not dropped]
    if len(non_dropped) > len(set(non_dropped)):
        raise KeyError('Column names are not unique')

    file_io = NMTRANDataIO(path_or_io, ignore_character)
    df = pd.read_table(
        file_io,
        sep=r' *, *| *[\t] *| +',
        na_filter=False,
        header=None,
        engine='python',
        quoting=3,
        dtype=object,
        index_col=False,
    )
    df = pharmpy.data.PharmDataFrame(df)

    diff_cols = len(df.columns) - len(colnames)
    if diff_cols > 0:
        df.columns = list(colnames) + [None] * diff_cols
        if not raw:
            # Remove unnamed columns
            df.drop(columns=[None], inplace=True)
    elif diff_cols < 0:
        if raw:
            df.columns = colnames[0 : len(df.columns)]
        else:
            warnings.warn("There are more columns in $INPUT than in the dataset")
            for i in range(abs(diff_cols)):  # Create empty columns.
                df[f'__{i}]'] = str(null_value)  # FIXME assure no name collisions here
            df.columns = colnames
    else:
        df.columns = colnames

    if coltypes is None:
        for label in df.columns:
            df.pharmpy.column_type[label] = infer_column_type(label)
    else:
        if len(coltypes) < len(df.columns):
            coltypes += [pharmpy.data.ColumnType.UNKNOWN] * (len(df.columns) - len(coltypes))
        df.pharmpy.column_type[list(df.columns)] = coltypes

    df = _filter_ignore_accept(df, ignore, accept, null_value)

    # if 'TIME' in df.columns:        # FIXME: Must handle synonyms
    #    try:
    #        id_label = df.pharmpy.id_label
    #        df = df.groupby(id_label).apply(_translate_nonmem_time_and_date)
    #    except KeyError:
    #        df.apply(_translate_nonmem_time_and_date)

    # FIXME: Do not remove dropped columns for now.
    # if drop:
    #    indices_to_drop = [i for i, x in enumerate(drop) if not x]
    #    df = df.iloc[:, indices_to_drop].copy()

    if not raw:
        parse_columns = [col for col, dropped in zip(df.columns, drop) if not dropped]
        # FIXME: This is instead of proper handling of these columns
        parse_columns = [
            x for x in parse_columns if x not in ['TIME', 'DATE', 'DAT1', 'DAT2', 'DAT3']
        ]
    for column in parse_columns:
        df[column] = df[column].apply(_convert_data_item, args=(str(null_value),))
    df = _make_ids_unique(df, parse_columns)

    # Make ID int if possible
    try:
        idcol = df.pharmpy.id_label
    except KeyError:
        pass
    else:
        df[idcol] = df[idcol].convert_dtypes()

    return df


def read_csv(path_or_io, raw=False, parse_columns=tuple()):
    """Read a csv with header into a PharmDataFrame"""
    if not raw:
        df = pd.read_csv(path_or_io)
    else:
        df = pd.read_csv(path_or_io, dtype=str)
        for col in parse_columns:
            df[col] = pd.to_numeric(df[col])

    df = pharmpy.data.PharmDataFrame(df)

    # Set name of PharmDataFrame in case we have a Pathlike input
    if isinstance(path_or_io, Path) or isinstance(path_or_io, str):
        path = Path(path_or_io)
        df.name = path.stem

    return df



import pytest

from pharmpy import Model
from pharmpy.modeling.help_functions import _as_integer, _get_etas


def test_get_etas(pheno_path, testdata):
    model = Model(pheno_path)

    etas = _get_etas(model, ['ETA(1)'])
    assert len(etas) == 1

    etas = _get_etas(model, ['ETA(1)', 'CL'], include_symbols=True)
    assert len(etas) == 1

    etas = _get_etas(model, ['ETA(1)', 'V'], include_symbols=True)
    assert len(etas) == 2

    with pytest.raises(KeyError):
        _get_etas(model, ['ETA(23)'])

    model = Model(testdata / 'nonmem' / 'pheno_block.mod')
    rvs = _get_etas(model, None)
    assert rvs[0].name == 'ETA(1)'

    model.parameters.fix = {'OMEGA(1,1)': True}
    rvs = _get_etas(model, None)
    assert rvs[0].name == 'ETA(2)'

    model = Model(testdata / 'nonmem' / 'pheno_block.mod')
    model.random_variables['ETA(1)'].level = 'IOV'
    rvs = _get_etas(model, None)
    assert rvs[0].name == 'ETA(2)'


def test_as_integer():
    n = _as_integer(1)
    assert isinstance(n, int) and n == 1
    n = _as_integer(1.0)
    assert isinstance(n, int) and n == 1
    with pytest.raises(TypeError):
        _as_integer(1.5)



from pharmpy import Model
from pharmpy.tools.modelfit.results import calculate_results


def test_modelfit(testdata):
    model = Model(testdata / 'nonmem' / 'pheno_real.mod')
    assert model


def test_aggregate(testdata):
    model = Model(testdata / 'nonmem' / 'pheno.mod')
    res = calculate_results([model, model])
    pe = res.parameter_estimates
    assert len(pe) == 2
    assert list(pe.index) == ['pheno', 'pheno']
    assert list(pe.columns) == ['THETA(1)', 'THETA(2)', 'OMEGA(1,1)', 'OMEGA(2,2)', 'SIGMA(1,1)']



from pharmpy.modeling.block_rvs import create_joint_distribution, split_joint_distribution
from pharmpy.modeling.common import (
    add_estimation_step,
    append_estimation_step_options,
    convert_model,
    copy_model,
    fix_parameters,
    fix_parameters_to,
    generate_model_code,
    get_model_covariates,
    load_example_model,
    print_model_symbols,
    read_model,
    read_model_from_database,
    read_model_from_string,
    remove_estimation_step,
    set_estimation_step,
    set_initial_estimates,
    set_name,
    unfix_parameters,
    unfix_parameters_to,
    write_model,
)
from pharmpy.modeling.covariate_effect import add_covariate_effect
from pharmpy.modeling.data import (
    get_number_of_individuals,
    get_number_of_observations,
    get_number_of_observations_per_individual,
    get_observations,
)
from pharmpy.modeling.error import (
    has_additive_error_model,
    has_combined_error_model,
    has_proportional_error_model,
    remove_error_model,
    set_additive_error_model,
    set_combined_error_model,
    set_dtbs_error_model,
    set_proportional_error_model,
    set_weighted_error_model,
    use_thetas_for_error_stdev,
)
from pharmpy.modeling.eta_additions import add_iiv, add_iov
from pharmpy.modeling.eta_transformations import (
    transform_etas_boxcox,
    transform_etas_john_draper,
    transform_etas_tdist,
)
from pharmpy.modeling.evaluation import evaluate_expression
from pharmpy.modeling.iiv_on_ruv import set_iiv_on_ruv
from pharmpy.modeling.ml import (
    predict_influential_individuals,
    predict_influential_outliers,
    predict_outliers,
)
from pharmpy.modeling.odes import (
    add_individual_parameter,
    add_lag_time,
    add_peripheral_compartment,
    has_zero_order_absorption,
    remove_lag_time,
    remove_peripheral_compartment,
    set_bolus_absorption,
    set_first_order_absorption,
    set_first_order_elimination,
    set_michaelis_menten_elimination,
    set_mixed_mm_fo_elimination,
    set_ode_solver,
    set_peripheral_compartments,
    set_seq_zo_fo_absorption,
    set_transit_compartments,
    set_zero_order_absorption,
    set_zero_order_elimination,
)
from pharmpy.modeling.parameter_sampling import (
    create_rng,
    sample_individual_estimates,
    sample_parameters_from_covariance_matrix,
    sample_parameters_uniformly,
)
from pharmpy.modeling.power_on_ruv import set_power_on_ruv
from pharmpy.modeling.remove_iiv import remove_iiv
from pharmpy.modeling.remove_iov import remove_iov
from pharmpy.modeling.results import (
    calculate_eta_shrinkage,
    calculate_individual_parameter_statistics,
    calculate_individual_shrinkage,
    calculate_pk_parameters_statistics,
    summarize_modelfit_results,
)
from pharmpy.modeling.run import create_results, fit, read_results, run_tool
from pharmpy.modeling.update_inits import update_inits

from .plots import plot_individual_predictions, plot_iofv_vs_iofv

# Remember to sort __all__ alphabetically for order in documentation
__all__ = [
    'add_covariate_effect',
    'add_estimation_step',
    'add_iiv',
    'add_individual_parameter',
    'add_iov',
    'add_lag_time',
    'add_peripheral_compartment',
    'append_estimation_step_options',
    'calculate_eta_shrinkage',
    'calculate_individual_parameter_statistics',
    'calculate_individual_shrinkage',
    'calculate_pk_parameters_statistics',
    'convert_model',
    'copy_model',
    'create_joint_distribution',
    'create_results',
    'create_rng',
    'evaluate_expression',
    'fit',
    'fix_parameters',
    'fix_parameters_to',
    'get_number_of_individuals',
    'get_number_of_observations',
    'get_number_of_observations_per_individual',
    'generate_model_code',
    'get_model_covariates',
    'get_observations',
    'has_additive_error_model',
    'has_combined_error_model',
    'has_proportional_error_model',
    'has_zero_order_absorption',
    'load_example_model',
    'plot_individual_predictions',
    'plot_iofv_vs_iofv',
    'predict_influential_individuals',
    'predict_influential_outliers',
    'predict_outliers',
    'print_model_symbols',
    'read_model',
    'read_model_from_database',
    'read_model_from_string',
    'read_results',
    'remove_error_model',
    'remove_estimation_step',
    'remove_iiv',
    'remove_iov',
    'remove_lag_time',
    'remove_peripheral_compartment',
    'run_tool',
    'sample_parameters_from_covariance_matrix',
    'sample_individual_estimates',
    'sample_parameters_uniformly',
    'set_additive_error_model',
    'set_bolus_absorption',
    'set_combined_error_model',
    'set_dtbs_error_model',
    'set_estimation_step',
    'set_first_order_absorption',
    'set_first_order_elimination',
    'set_iiv_on_ruv',
    'set_initial_estimates',
    'set_michaelis_menten_elimination',
    'set_mixed_mm_fo_elimination',
    'set_name',
    'set_ode_solver',
    'set_peripheral_compartments',
    'set_power_on_ruv',
    'set_proportional_error_model',
    'set_seq_zo_fo_absorption',
    'set_transit_compartments',
    'set_weighted_error_model',
    'set_zero_order_absorption',
    'set_zero_order_elimination',
    'split_joint_distribution',
    'summarize_modelfit_results',
    'transform_etas_boxcox',
    'transform_etas_john_draper',
    'transform_etas_tdist',
    'unfix_parameters',
    'unfix_parameters_to',
    'update_inits',
    'use_thetas_for_error_stdev',
    'write_model',
]
